import os
import numpy as np
from pprint import pp
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl
from itertools import starmap

from mcfnmr.demodata import (
    mixNamesOLDB,
    expectedConcentrationFactors,
    compoundLongNames,
    correctShortCompoundLabels,
    scanNumbersMixes, textbook_abbreviations,
)
from mcfnmr.demodata.loading import loadOLDBCompoundLib, buildHMDBID2NameMap
from mcfnmr.routines.utils import (
    totalWeights,
    figs_dir_MM,
    figs_dir_oldb,
    filterDF,
    figs_dir,
)
from mcfnmr.config import DEBUG, OUTDIR


def plotPredictionDiagonals(dfCompounds_dict, ar, setup=None, show=False):
    if setup:
        isoFit, incrFit = setup
    dfs = list()
    for k, d in dfCompounds_dict.items():
        if setup and setup != k:
            continue
        for mixID, df in d.items():
            g = df.groupby("assignment radius")
            df = df.loc[g.groups[ar]]
            if setup:
                g = df.groupby("incremental fit")
                df = df.loc[g.groups[incrFit]]
                g = df.groupby("isolated fit")
                df = df.loc[g.groups[isoFit]]
            dfs.append(df)
    df = pd.concat(dfs)

    make_setup_str = lambda iso, incr: ("iso-" if iso else "joint-") + (
        "incr" if incr else "onepass"
    )
    m = list(starmap(make_setup_str, zip(df["isolated fit"], df["incremental fit"])))
    df["setup"] = list(m)
    df["predicted"] = df["concentration factor"] * 3
    df["expected"] = np.round(df["expected concentration factor"] * 3, 10)
    df["signal-to-noise"] = df["expected"] * np.sqrt(
        [scanNumbersMixes[targetID] for targetID in df["target"]]
    )
    
    if setup:
        df.to_csv(figs_dir / f"quantitative_predictions_compounds_setup{setup}.csv", sep=",")
    else:
        df.to_csv(figs_dir / "quantitative_predictions_compounds.csv", sep=",")
    
    # Jitter out expected in Y-axis to get
    # a better view on point density
    jfac = 0.06
    rng = np.random.default_rng(123)
    jitter = 1 + 2 * jfac * rng.random(len(df["expected"])) - jfac
    df["expected (jittered)"] = df["expected"] * jitter

    fig, ax = plt.subplots(layout="constrained", figsize=(4.0, 4.0))

    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_ylim([0.01, 50])
    ax.set_xlim([0.1, 40])
    if setup:
        sns.stripplot(df, y="predicted", x="expected", native_scale=True, s=3, ax=ax, zorder=0)
        sns.boxplot(df, y="predicted", x="expected", native_scale=True, 
                    fliersize=0, boxprops=dict(alpha=0.75), ax=ax, zorder=1)
    else:
        sns.stripplot(df, y="predicted", x="expected", hue="setup", native_scale=True, s=3, ax=ax, zorder=0)
        sns.boxplot(df, y="predicted", x="expected", hue="setup", native_scale=True, 
                    fliersize=0, boxprops=dict(alpha=0.75), ax=ax, zorder=1)
    ax.plot([0, 50], [0, 50], "k--", lw=0.5, zorder=-2)
    # ax.set_ylim([0.0, 50])
    # ax.set_xlim([0.0, 40])
    ax.set_title("predicted vs expected concentrations at ar=%g (excl. exp.=0)" % ar)

    if setup:
        setupstr = df["setup"].iloc[0]
        figname = (
            figs_dir / f"predicted_vs_expected_concentrations_hue_by_S2N_{setupstr}.svg"
        )
    else:
        figname = figs_dir / "predicted_vs_expected_concentrations_hue_by_setup.svg"
    fig.savefig(figname)
    print(f"Saved figure as '{figname}'")
    
    plot_single_compounds = True
    if plot_single_compounds:
        R2s = {}
        loglog = True
        ext = "svg"
        df.index = list(range(len(df)))
        if setup is None or setup == (False, True):
            if loglog:
                figsubdir = figs_dir / "concentration_prediction_per_cpd_loglog"
            else:
                figsubdir = figs_dir / "concentration_prediction_per_cpd"
            if not figsubdir.exists():
                os.makedirs(figsubdir)
            
            gg = df.groupby("compound").groups
            for c in sorted(set(df["compound"])):
                ix = gg[c]
                if setup:
                    figname = figsubdir / f"expected_predicted_{c}_setup{setup}.{ext}"
                else:
                    figname = figsubdir / f"expected_predicted_{c}.{ext}"
                dfc = df.iloc[ix,:]
                fig, ax = plt.subplots(figsize = (2.0,2.0), layout="constrained")
                sns.scatterplot(dfc, y="predicted", x="expected", hue="setup", s=10, ax=ax, zorder=0)
                if loglog:
                    ax.set_yscale("log")
                    ax.set_xscale("log")
                xlim, ylim = ax.get_xlim(), ax.get_ylim()
                lims = (min(xlim[0], ylim[0]), max(xlim[1], ylim[1]))
                lims = (0.01, lims[1]) # for indentical xlims for manuscript figures
                ax.set_xlim(lims), ax.set_ylim(lims)
                ax.plot(lims,lims, "k--", lw=0.5, zorder=-2)
                
                # Fit linear regression and plot regression curve
                import statsmodels.api as sm
                if loglog:
                    mod = sm.OLS(dfc["predicted"], dfc["expected"])
                else:
                    # Fit with intercept
                    dfc["Intercept"] = 1.0
                    mod = sm.OLS(dfc["predicted"], dfc[["Intercept", "expected"]])
                res = mod.fit()
                print(res.summary())
                a = 0.0 if loglog else res.params["Intercept"]
                b = res.params["expected"]
                R2 = res.rsquared
                R2s[compoundLongNames[c]]=R2
                avg_err = np.nanmean(np.abs(dfc["predicted"] - dfc["expected"])/dfc["expected"])
                # xspan = np.exp(np.linspace(*np.log(lims), 101))
                xspan = np.linspace(*lims, 101)
                ax.plot(xspan, a+b*xspan, color="red", zorder=-1)
                ax.set_title(textbook_abbreviations[compoundLongNames[c]] + f" ($R^2={R2:g}$, avg.err={avg_err:g})")
                fig.savefig(figname)
                print(f"Saved figure '{figname}'")
                # plt.show()
                plt.close(fig)
            print("\nR2S:")
            pp(R2s)
            
    if show:
        plt.show()


def plotAssignmentStatsOLDB(
    dfCompounds,
    dfTargets,
    libID,
    targetID,
    showCompounds=None,
    markCompounds=[],
    outdir=None,
    show=False,
    setupStr="",
):
    assert outdir is not None
    targetName = mixNamesOLDB.get(targetID, targetID)
    if markCompounds is None:
        markCompounds = []
    arSpan = sorted(set(dfCompounds["assignment radius"]))

    # gathers info during preparing plots (return of this function)
    stats = {}

    errorStats = {
        "mean relative error": [],
        "percentage mid range": [],
        "std(err)": [],
        "assignment radius": [],
        "mixture": [],
        "matching": [],
        "false positives": [],
    }

    # For each assignment radius, plot the single compound stats
    nrows = len(arSpan)

    # Setup main figure: selected compound's match vs expectation
    # and error distributions for different ar
    plt.style.use("bmh")
    fig_main = plt.figure(figsize=(3.3, 2.3))  # , layout="tight")
    fig, fig2 = fig_main.subfigures(nrows=1, ncols=2, width_ratios=(1, 1))
    # fig.get_layout_engine().set(hspace=10, wspace=10)
    for i in range(nrows):
        ax1 = fig.add_subplot(nrows, 1, i + 1)
        ax2 = fig2.add_subplot(nrows, 1, i + 1)
        if i != nrows - 1:
            ax1.set_xlabel("")
            ax2.set_xlabel("")
            ax1.set_xticklabels([])
            ax2.set_xticklabels([])

    axes, axes2 = fig.get_axes(), fig2.get_axes()
    fig.suptitle("Assignment (%s)" % (setupStr))
    fig2.suptitle("Prediction error (%s)" % (setupStr))

    # Extra plot for the four compounds with higher concetrations
    selectedCompounds = ["Bnz3", "Gluc", "Pim", "Tyr"]
    nExpect = len(selectedCompounds)

    if nrows == 0:
        raise Exception("plotAssignmentStatsOLDB(): error")

    fig1, axes1 = plt.subplots(
        ncols=nrows, figsize=(2 * nrows, min(10, nExpect * 0.75))
    )
    fig1.suptitle("Assignment present (%s)" % (setupStr))

    filters = {"lib": libID, "target": targetName}
    dfC = filterDF(dfCompounds, filters)
    dfT = filterDF(dfTargets, filters)
    totalWeightY = dfT["originalWeightY"].iloc[0]
    lib = loadOLDBCompoundLib()
    cweights = totalWeights(lib)

    print(f"\n\n#### {setupStr} ####")
    # Ensure axes are iterable
    if nrows == 1:
        axes1 = [axes1]

    for ar, ax, ax1, ax2 in zip(arSpan, axes, axes1, axes2):
        dfr = dfC.loc[dfC["assignment radius"] == ar].copy()
        dfr.loc[:, "expected"] = [
            expectedConcentrationFactors[targetID][compoundLongNames.get(c, c)]
            for c in dfr["compound"]
        ]
        totalDetected = sum(
            dfr["concentration factor"]
            * np.array([cweights[compoundLongNames.get(c, c)] for c in dfr["compound"]])
        )
        totalExpected = sum(
            dfr["expected"]
            * np.array([cweights[compoundLongNames.get(c, c)] for c in dfr["compound"]])
        )

        print(f"\nAssignment Radius {ar}:")
        print(f"    totalWeightY = {totalWeightY}")
        print(f"    totalExpected = {totalExpected}")
        print(f"    totalDetected = {totalDetected}")

        for c, cf, ex in zip(
            dfr["compound"], dfr["concentration factor"], dfr["expected"]
        ):
            if ex > 0:
                print(f"{c}: cf/ex = %g" % (cf / ex))

        dfr.loc[:, "error"] = (
            dfr.loc[:, "concentration factor"] - dfr.loc[:, "expected"]
        )
        dfr.loc[:, "rel. error"] = dfr.loc[:, "error"] / dfr.loc[:, "expected"]
        infrows = dfr["rel. error"] == np.inf
        dfr.loc[infrows, "rel. error"] = np.nan
        uniform_matchtype = (
            len(set(dfr["isolated fit"])) == 1 and len(set(dfr["incremental fit"])) == 1
        )
        if uniform_matchtype:
            isoFitStr = "indep." if list(set(dfr["isolated fit"]))[0] else "joint"
            incrFitStr = (
                " incr. fit" if list(set(dfr["incremental fit"]))[0] else " fit"
            )
            match_type = f"{isoFitStr}{incrFitStr}"
        else:
            match_type = "various"

        # Calculate mean relative error and midrange percentage
        # for present compounds
        dfr_present = dfr[dfr["expected"] > 0]
        std = np.nanstd(dfr_present["rel. error"])
        meanErr = np.nanmean(dfr_present["rel. error"])
        meanAbsErr = np.nanmean(np.abs(dfr_present["rel. error"]))
        pctInMidrange = (
            100
            * np.count_nonzero(
                np.logical_and(
                    dfr_present["rel. error"] > -0.5, dfr_present["rel. error"] < 1.0
                )
            )
            / len(dfr_present["rel. error"])
        )

        # Check if any compounds which are not present have benn detected
        fp_thresh = np.max(dfr["expected"]) * 0.01
        FP = dfr["compound"][
            (dfr["expected"] == 0) & (dfr["concentration factor"] > fp_thresh)
        ]
        if len(FP) > 0:
            print(f"Found false positive detection for {targetID}: {FP}")
        errorStats["false positives"].append(len(FP))
        errorStats["mean relative error"].append(meanAbsErr)
        errorStats["percentage mid range"].append(pctInMidrange)
        errorStats["std(err)"].append(std)
        errorStats["assignment radius"].append(ar)
        errorStats["mixture"].append(targetID)
        errorStats["matching"].append(match_type)

        dfr.loc[:, "in mix"] = np.isin(dfr.loc[:, "compound"], markCompounds)
        dfr_expected = dfr.loc[np.isin(dfr.loc[:, "compound"], selectedCompounds)]
        if len(dfr["assigned"]) == 0:
            continue

        # Filter data to only show thespecified set of compounds
        if showCompounds is not None:
            ix = [c in showCompounds for c in dfr["compound"]]
            dfrshow = dfr[ix]
        else:
            dfrshow = dfr

        shortlabs = correctShortCompoundLabels(dfrshow["compound"])
        dfrshow.loc[:, "compound"] = shortlabs
        shortlabs = correctShortCompoundLabels(dfr_expected["compound"])
        dfr_expected.loc[:, "compound"] = shortlabs
        dodge = True

        sns.barplot(
            dfrshow,
            x="compound",
            y="concentration factor",
            color="xkcd:grass",
            dodge=dodge,
            ax=ax,
        )
        sns.barplot(
            dfrshow,
            x="compound",
            y="expected",
            color="xkcd:yellow",
            dodge=dodge,
            alpha=0.5,
            ax=ax,
        )

        if len(markCompounds) > 0:
            for label in ax.get_yticklabels():
                col = "xkcd:green" if label.get_text() in markCompounds else "xkcd:red"
                label.set_color(col)
        ylim = (0.0, 1.5 * max(dfr["expected"]))
        ax.set_ylim(ylim)
        ax.text(0.2, ylim[1] * 0.9, "$r = %g$" % ar, va="top")
        if ax == axes[-1]:
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(ax.get_xticklabels(), rotation=60)
            ax.set_ylabel("")
        if ax == axes[0]:
            ax.set_ylabel("")

        ax1.set_title("$r = %g$" % ar)
        sns.barplot(
            dfr_expected,
            x="compound",
            y="concentration factor",
            color="xkcd:grass",
            dodge=dodge,
            ax=ax1,
        )
        sns.barplot(
            dfr_expected,
            x="compound",
            y="expected",
            color="xkcd:yellow",
            dodge=dodge,
            alpha=0.5,
            ax=ax1,
        )
        ax1.set_ylim((0.0, 3 * max(dfr["expected"])))

        ax1.set_xticks(ax1.get_xticks())
        ax1.set_xticklabels(ax1.get_xticklabels(), rotation=60)

        pal = sns.color_palette("viridis", n_colors=5)
        histRange = [-1.1, 2.5]
        nbins = 18

        if len(set(dfr["expected"])) > 1:
            ylims = [0, 25]
            ho = (max(dfr["expected"]), min(dfr["expected"]))
            sns.histplot(
                dfr,
                x="rel. error",
                hue="expected",
                hue_order=ho,
                multiple="layer",
                stat="percent",
                common_norm=False,
                bins=nbins,
                binrange=histRange,
                ax=ax2,
                palette=[pal[0], pal[3]],
            )
        else:
            ylims = [0, 12]
            sns.histplot(
                dfr,
                x="rel. error",
                stat="count",
                bins=nbins,
                binrange=histRange,
                color=pal[0],
                ax=ax2,
            )
        ax2.set_ylim(ylims)
        ax2.plot([meanErr, meanErr], ylims, ls="--", lw=1.0, color=pal[0], label="mean")
        ylims = ax2.get_ylim()
        ax2.set_ylim(ylims)
        ax2.plot([-0.5, -0.5], ylims, ls="--", lw=1.0, color="xkcd:salmon")
        ax2.plot([1, 1], ylims, ls="--", lw=1.0, color="xkcd:salmon")
        ax2title = "$r = %g$, std=%g, %s in [1/2, 2]: %.1f" % (
            ar,
            std,
            "%",
            pctInMidrange,
        )
        ax2.set_title(ax2title)

    stats["errors"] = errorStats

    figname = outdir / f"assignment and prediction_err {setupStr}.png"
    fig_main.savefig(figname)
    figname = outdir / f"assignment and prediction_err {setupStr}.svg"
    fig_main.savefig(figname)
    print(f"Saved fig {figname}")
    if not show:
        plt.close(fig_main)

    fig1.tight_layout()
    figname = outdir / f"compoundAssignment to expected {setupStr}.png"
    fig1.savefig(figname)
    figname = outdir / f"compoundAssignment to expected {setupStr}.svg"
    fig1.savefig(figname)
    print(f"Saved fig {figname}")
    if show:
        plt.show()
    else:
        plt.close(fig1)

    return stats


def plotErrorStats(data, figname, show=True):
    """
    For different experimental series, plot assignment radius against
    mean rel. error and percentage mid range
    """

    # Experimental series
    series = {
        # all cpds:
        "all": ["Ia_01", "Ib_01", "Ic_01"],
        # 4 cpds:
        "4 cpds": ["IIa_01", "IIb_01", "IIc_01"],
        # all + 4 cpds:
        "all+4": ["IIIa_01", "IIIb_01", "IIIc_01"],
    }

    df = pd.DataFrame(data)

    mix_descriptions = {
        "Ia_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=3\ mM$,",
        "Ib_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=0.3\ mM$,",
        "Ic_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=0.03\ mM$,",
        "IIa_01": r"$N_{\mathrm{mix}}=4$,\n$c_k^\ast=30\ mM$,",
        "IIb_01": r"$N_{\mathrm{mix}}=4$,\n$c_k^\ast=3\ mM$,",
        "IIc_01": r"$N_{\mathrm{mix}}=4$,\n$c_k^\ast=0.3\ mM$,",
        "IIIa_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=\{0.15, 15.15\}\ mM$,",
        "IIIb_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=\{0.15, 1.65\}\ mM$,",
        "IIIc_01": r"$N_{\mathrm{mix}}=34$,\n$c_k^\ast=\{0.15, 0.30\}\ mM$,",
    }
    for k in mix_descriptions.keys():
        nScans = scanNumbersMixes[k]
        mix_descriptions[k] += "\n" + r"$\mathrm{NS}=" + ("%d$" % nScans)

    ylims = {
        "all": dict(midrange=[0, 105], relerr=[0, 2.2]),
        "4 cpds": dict(midrange=[0, 105], relerr=[0, 2.2], fps=[0, 34]),
        "all+4": dict(midrange=[0, 105], relerr=[0, 2.2]),
    }

    plt.style.use("bmh")
    fig = plt.figure(figsize=(10, 12))
    figs = fig.subfigures(3, 1, wspace=0.01)

    for s, f in zip(series, figs):
        four_cpds = False
        mixes = series[s]
        f.suptitle(s + f" ({mixes})")
        # reference axes for sharing y-axis
        for i, m in enumerate(mixes):
            dfm = df[df["mixture"] == m]

            print(f"Mix {m}:")
            for h in sorted(set(dfm["matching"])):
                dfh = dfm[dfm["matching"] == h]
                minErrIx = np.argmin(dfh["mean relative error"])
                minErrAr = dfh["assignment radius"].iloc[minErrIx]
                minErr = dfh["mean relative error"].iloc[minErrIx]
                maxMidrngIx = np.argmax(dfh["percentage mid range"])
                maxMidrngAr = dfh["assignment radius"].iloc[maxMidrngIx]
                maxMidrng = dfh["percentage mid range"].iloc[maxMidrngIx]
                if DEBUG > 0:
                    print(f"  setup {h}:")
                    print("    min rel err: %g (at r=%g)" % (minErr, minErrAr))
                    print("    max mid rng: %g (at r=%g)" % (maxMidrng, maxMidrngAr))

            ax1 = f.add_subplot(
                2 + four_cpds, len(mixes), i + 1
            )  # , sharey=share_ax_err)
            sns.lineplot(
                dfm,
                x="assignment radius",
                y="mean relative error",
                hue="matching",
                ax=ax1,
            )
            if m in ["0203_01", "0203_01", "0207_01"]:
                ax1.text(0.21, 1.05, mix_descriptions[m], va="top")
            else:
                ax1.text(0.01, 2.05, mix_descriptions[m], va="top")
            ax1.legend(loc=2).remove()
            ax1.set_ylim(ylims[s]["relerr"])
            ax1.set_xlim((0, 0.5))

            ax2 = f.add_subplot(
                2 + four_cpds, len(mixes), len(mixes) + i + 1
            )  # , sharey=share_ax_mid)
            sns.lineplot(
                dfm,
                x="assignment radius",
                y="percentage mid range",
                hue="matching",
                ax=ax2,
            )
            ax2.set_ylim(ylims[s]["midrange"])
            ax2.set_xlim((0, 0.5))

            if four_cpds:
                ax3 = f.add_subplot(
                    2 + four_cpds, len(mixes), 2 * len(mixes) + i + 1
                )  # , sharey=share_ax_mid)
                sns.lineplot(
                    dfm,
                    x="assignment radius",
                    y="false positives",
                    hue="matching",
                    ax=ax3,
                )
                ax3.set_ylim(ylims[s]["fps"])
                ax3.legend(loc=2).remove()
                ax3.set_xlim((0, 0.5))

            # Remove unnecessary labels
            ax1.set_xlabel("")
            ax1.set_xticklabels("")
            if i != 0:
                ax1.set_ylabel("")
                ax2.set_ylabel("")
                ax1.set_yticklabels("")
                ax2.set_yticklabels("")
            if i != len(mixes) - 1 or s != "all":
                ax2.legend(loc=2).remove()

    fn = figs_dir_oldb / f"{figname}.png"
    fig.savefig(fn)
    fn = figs_dir_oldb / f"{figname}.svg"
    fig.savefig(fn)
    print(f"Saved fig {fn}")
    if show:
        plt.show()
    else:
        plt.close(fig)


def plotTimings(results, setup, mixIDs, loglog=True, show=True):
    fig, ax = plt.subplots(layout="constrained", figsize=(4, 3))
    # fig, ax = plt.subplots(layout="constrained", figsize=(6,4))
    mcfTimes = {}
    colors = ["xkcd:blue", "xkcd:orange", "xkcd:green"]
    for i, mixID in enumerate(mixIDs):
        timings = {ar: res[mixID].timings for ar, res in results.items()}
        arSpan = np.array(sorted(timings))
        if mixID == mixIDs[0]:
            ax.plot(arSpan * arSpan, np.zeros_like(arSpan), "k--", lw=1)
        if setup == "setup D":
            mcfTimesSecs = [timings[ar]["minCostFlow"] for ar in arSpan]
            mcfTimes[mixID] = mcfTimesSecs
        else:
            mcfTimesMins = [timings[ar]["minCostFlow"] / 60 for ar in arSpan]
            mcfTimes[mixID] = mcfTimesMins
        if loglog:
            ax.plot(arSpan, mcfTimes[mixID], marker=".", label=mixID, color=colors[i])
        else:
            ax.plot(
                arSpan * arSpan,
                mcfTimes[mixID],
                marker=".",
                label=mixID,
                color=colors[i],
            )
    if loglog:
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel("assignment radius $r$")
        ax.set_xlim(arSpan[[0, -1]])
    else:
        ticks = arSpan * arSpan
        labels = ["$%g\\times%g$" % (ar, ar) for ar in arSpan]
        ax.set_xticks(ticks, labels, rotation=45)
        ax.set_xlabel("assignment area $r^2$")
        ax.set_xlim((arSpan * arSpan)[[0, -1]])
    ax.set_title(f"Timings linprog() {setup}")
    if setup == "setup D":
        ax.set_ylabel("time [secs]")
    else:
        ax.set_ylabel("time [mins]")
    ax.legend()

    print("timings:")
    pp(mcfTimes)

    # Test asymptotic, effective complexity
    asr_ix = [-5, -1]  # asymptotic range
    print("asymptotic range:", arSpan[asr_ix])
    logt = {mixID: np.log(mcfTimes[mixID])[asr_ix] for mixID in mixIDs}
    log_asr = np.log(arSpan[asr_ix])
    slopes = {
        mixID: (logt[mixID][1] - logt[mixID][0]) / (log_asr[1] - log_asr[0])
        for mixID in mixIDs
    }
    print("log-log slopes (time~ar^alpha):")
    pp(slopes)

    if loglog:
        for i, mixID in enumerate(mixIDs):
            ax.plot(
                arSpan[asr_ix],
                np.array(mcfTimes[mixID])[asr_ix],
                ls="--",
                color=colors[i],
            )

    logstr = "_loglog" if loglog else ""
    figname = figs_dir / f"Timings_{setup}{logstr}.svg"
    fig.savefig(figname)
    print(f"Saved fig '{figname}'")

    if show:
        plt.show()


def plotPerformanceScanResults(
    scanResults,
    title=None,
    logAR=True,
    plotPcolor=True,
    plotContour=False,
    outdir=None,
    show=False,
    plotDetectionsInsteadOfPrecision=False,
    verb=1,
    figsize=None,
    plot_max_marker=True,
):

    arSpan = sorted(scanResults.keys())
    thSpan = sorted(scanResults[arSpan[0]].keys())
    nar, nth = len(arSpan), len(thSpan)

    # Collect data into matrices
    fscores = np.zeros((nar, nth))
    recalls = np.zeros((nar, nth))
    precisions = np.zeros((nar, nth))
    detections = np.zeros((nar, nth))
    max_ix, maxfscore = (0, 0), 0.0
    for i, ar in enumerate(arSpan):
        if verb > 0:
            print("\nar=%g" % ar)
        for j, th in enumerate(thSpan):
            if verb > 0:
                print("th=%g" % th)
            stats = scanResults[ar][th]
            mixIDs = sorted(stats.keys())
            data = [stats[mixID]["classification"] for mixID in mixIDs]
            fscores[i, j] = np.mean(np.array([d["fscore"] for d in data]))
            recalls[i, j] = np.mean(np.array([d["recall"] for d in data]))
            precisions[i, j] = np.mean(np.array([d["precision"] for d in data]))
            detections[i, j] = np.sum(
                np.array([len(d["TP"]) + len(d["FP"]) for d in data])
            )
            if verb > 0:
                print("fscore=%g" % fscores[i, j])
                print("recall=%g" % recalls[i, j])
                print("precision=%g" % precisions[i, j])

            if fscores[i, j] > maxfscore:
                max_ix = (i, j)
                maxfscore = fscores[i, j]
    armax, thmax = arSpan[max_ix[0]], thSpan[max_ix[1]]
    print("\n### Maximal fscore: %g (at: ar=%g, th=%g)" % (maxfscore, armax, thmax))
    print(
        "    recall: %g, precision: %g"
        % (recalls[max_ix[0], max_ix[1]], precisions[max_ix[0], max_ix[1]])
    )

    maxstats = dict(
        [
            (mixID, scanResults[armax][thmax][mixID]["classification"])
            for mixID in mixIDs
        ]
    )
    for mixID, data in maxstats.items():
        print("  mix %s:" % mixID)
        print("      recall: %g" % data["recall"])
        print("      precision: %g" % data["precision"])
        print("      fscore: %g" % data["fscore"])

    # Plot contour
    X, Y = np.meshgrid(thSpan, arSpan)
    if figsize is None:
        # figsize = (5,2.) # For main text 2x2
        figsize = (8, 2.7)  # For supplement 1x
    fig, axes = plt.subplots(ncols=3, sharey=True, figsize=figsize)
    fig.suptitle(title)
    axes[0].set_ylabel("assignment radius")
    if plotDetectionsInsteadOfPrecision:
        labs = ["F1 score", "recall", "detections"]
        fields = [fscores, recalls, detections]
    else:
        labs = ["F1 score", "recall", "precision"]
        fields = [fscores, recalls, precisions]
    for ax, vals, lab in zip(axes, fields, labs):
        if plotPcolor:
            if lab == "detections":
                cf = ax.pcolormesh(X, Y, vals, cmap="magma")
            else:
                cf = ax.pcolormesh(X, Y, vals, vmin=0.0, vmax=1.0, cmap="magma")
            if plotContour:
                ax.contour(
                    X,
                    Y,
                    vals,
                    vmin=0.0,
                    vmax=1.0,
                    levels=np.linspace(0.0, 1.0, 11),
                    zorder=100,
                    linewidths=0.5,
                    alpha=0.5,
                    cmap="Greys",
                )
        else:
            cf = ax.contourf(
                X,
                Y,
                vals,
                vmin=0.0,
                vmax=1.0,
                cmap="inferno",
                levels=np.linspace(0.0, 1.0, 11),
            )
        
        if plot_max_marker:
            ax.scatter(thmax, armax, s=15, c="xkcd:red", marker="+", linewidths=0.8)
        ax.set_title(lab)
        ax.set_xlabel("detection threshold")
        # ax.grid(color="#eeeeee", zorder=1)
        if logAR:
            ax.set_yscale("log")
    # fig.colorbar(cf)
    plt.tight_layout()

    if outdir is None:
        outdir = figs_dir_MM
    figname = outdir / (title + ".png")
    fig.savefig(figname, dpi=600)
    print(f"Saved fig '{figname}'")
    figname = outdir / (title + ".svg")
    fig.savefig(figname, dpi=600)
    print(f"Saved fig '{figname}'")

    if show:
        plt.show()
    else:
        plt.close(fig)


def plotSlices(
    scanResults, ar_fix, th_fix, figname, figsize, report=True, max_th=None, show=True
):
    scores = scanResults
    mixID = sorted(scanResults[ar_fix][th_fix].keys())[0]
    plt.style.use("bmh")
    arSpan = sorted(scanResults.keys())
    thSpan = sorted(scanResults[arSpan[0]].keys())

    fig, axes = plt.subplots(nrows=2, layout="constrained", figsize=figsize)
    data = {th: scanResults[ar_fix][th][mixID]["classification"] for th in thSpan}
    fscores = [data[th]["fscore"] for th in thSpan]
    # recall    = [data[th]["recall"] for th in thSpan]
    # precision = [data[th]["precision"] for th in thSpan]
    TPs = [len(data[th]["TP"]) for th in thSpan]
    FPs = [len(data[th]["FP"]) for th in thSpan]
    FNs = [len(data[th]["FN"]) for th in thSpan]
    ax2 = axes[0]
    ax2.plot(thSpan, TPs, label="confirmed")
    ax2.plot(thSpan, FNs, label="undetected")
    ax2.plot(thSpan, FPs, label="unconfirmed")
    ax2.set_title("Slice at $r=%g$" % ar_fix)
    ax2.set_xlabel("detection threshold")
    ax2.set_ylabel("compounds")
    ax2.set_ylim((-0.5, 50))
    if max_th:
        ax2.set_xlim((thSpan[0], max_th))
    else:
        ax2.set_xlim((thSpan[0], thSpan[-1]))

    print("\nMaximal 'fscore' within th-slice at ar=%g: %g" % (ar_fix, max(fscores)))

    data = {ar: scanResults[ar][th_fix][mixID]["classification"] for ar in arSpan}
    fscores = [data[ar]["fscore"] for ar in arSpan]
    # recall    = [data[ar]["recall"] for ar in arSpan]
    # precision = [data[ar]["precision"] for ar in arSpan]
    TPs = [len(data[ar]["TP"]) for ar in arSpan]
    FPs = [len(data[ar]["FP"]) for ar in arSpan]
    FNs = [len(data[ar]["FN"]) for ar in arSpan]
    ax2 = axes[1]
    ax2.plot(arSpan, TPs)
    ax2.plot(arSpan, FNs)
    ax2.plot(arSpan, FPs)
    fig.legend()
    ax2.set_title("Slice at $\\vartheta=%g$" % th_fix)
    ax2.set_xlabel("assignment radius")
    ax2.set_ylabel("compounds")
    ax2.set_ylim((-0.5, 50))
    ax2.set_xlim((arSpan[0], arSpan[-1]))

    print("Maximal 'fscore' within ar-slice at th=%g: %g\n" % (th_fix, max(fscores)))

    if figname:
        fig.savefig(figname)
        print(f"Saved figure '{figname}'")

    if report:
        id2name = buildHMDBID2NameMap(verb=0)
        print("\n# plotSlices() detection report:")
        prev = {}
        for th in reversed(thSpan):
            data = scanResults[ar_fix][th][mixID]["classification"]
            print(f"\nAt ar={ar_fix:g}, th={th:g}:")
            print(f"   # TP: {len(data['TP'])}")
            print(f"   # FP: {len(data['FP'])}")
            print(f"   # FN: {len(data['FN'])}")
            newTPs = set(data["TP"]).difference(prev.get("TP", []))
            newFPs = set(data["FP"]).difference(prev.get("FP", []))
            print(f"new TPs:")
            for ID in newTPs:
                print(f"  {ID} ({id2name[ID]}),")
            print(f"new FPs:")
            for ID in newFPs:
                print(f"  {ID} ({id2name[ID]}),")
            prev = data

    if show:
        plt.show()
    else:
        plt.close(fig)



import os
import pickle
from timeit import default_timer
import numpy as np
import networkx as nx
from scipy.optimize import linprog
from scipy.sparse import lil_array

from mcfnmr.config import (
    DEFAULT_SCALE_C,
    DEFAULT_SCALE_H,
    MAX_ABSORPTON_COST,
    OUTDIR,
    DEBUG,
)
from mcfnmr.utils.geometry import load_regions, getIndicesWithin
from mcfnmr.utils.pointspectrum import detect_grid


class MCFResult(object):
    def __init__(self, total_target_weight, stats, absorption_cost, compounds, timings):
        self.originalWeightY = total_target_weight
        self.absorptionCost = absorption_cost
        self.timings = timings
        self.compounds = compounds
        compoundIDs = sorted(compounds)

        self.nPeaksConsidered = stats.get("nPeaksConsidered")
        self.inflowsY = stats.get("inflowsY")
        self.residualsY = stats.get("residualsY")
        self.residualsX = stats.get("residualsX")
        self.assigned = stats.get("flowsC2Y")
        total_weights = {
            c: compounds[c]["total_weight"] if compounds[c]["total_weight"] > 0 else 1.0
            for c in compoundIDs
        }
        self.assigned_relative = {
            c: self.assigned[i] / total_weights[c] for i, c in enumerate(compoundIDs)
        }
        self.assignmentCosts = stats.get("costsC2Y")
        self.unknownFlow = stats.get("unknownFlowC")
        self.specificCosts = stats.get("specificCostsC")
        self.inflowsC = stats.get("inflowsC")
        self.absorbedFlow = stats.get("absorbedFlow")
        self.totalAbsorption = np.sum(self.absorbedFlow)
        self.totalAssignedFlow = np.sum(self.assigned)
        self.totalAbsorptionCost = self.totalAbsorption * absorption_cost
        self.totalAssignmentCost = sum(self.assignmentCosts)
        self.totalCost = self.totalAbsorptionCost + self.totalAssignmentCost

        print("Total flows:")
        print("   absorbed: %g" % self.totalAbsorption)
        print("   assigned: %g" % self.totalAssignedFlow)
        print("Total costs:")
        print(" assignment: %g" % self.totalAssignmentCost)
        print(" absorption: %g" % self.totalAbsorptionCost)
        print("      total: %g" % self.totalCost)


def serialize(lib):
    compounds = sorted(lib.keys())
    ixs, all_weights, all_coords = {}, [], []
    current_ix = 0
    for compound_name in compounds:
        spectrum = lib[compound_name]
        if np.any(spectrum.weights < 0):
            raise Exception(f"Negative weight for '{spectrum.name}'")
        ixs[compound_name] = np.arange(current_ix, current_ix + len(spectrum.weights))
        all_weights.extend(spectrum.weights)
        all_coords.extend(spectrum.coords)
        current_ix += len(spectrum.weights)
    return ixs, np.array(all_weights), np.array(all_coords)


def euclidean_dists(xCoords, yCoords, scales, assignment_radius):
    """
    Returns two lists (distList, indexList) of length nx containing distances
    and corresponding indices of y-coords, where x and y distances are weighted
    according to the given scalings. Only points within the given assignment radius
    are included into the lists.
    """
    assert yCoords.shape[1] == xCoords.shape[1]
    nx, ny, ndim = xCoords.shape[0], yCoords.shape[0], xCoords.shape[1]
    assert len(scales) == ndim
    
    # Try to deduce whether target is a grid spectrum
    res = detect_grid(yCoords)
    griddedY = res[0] is not None
    if DEBUG and not griddedY:
        print(f"Could not deduce grid for spectrum.\n   ({res[1]})")
    else:
        _, _, indexer = res
    
    distList, indexList = [], []
    for i in range(nx):
        x = xCoords[i]
        if griddedY:
            # Determine indices in rectangle  corresponding to quadrat of side 
            # length assignment_radius in rescaled coords
            rx, ry = assignment_radius*scales[0], assignment_radius*scales[1]
            # xspan, yspan indices
            ixs = indexer.getNeighs(x, rx, ry)
            
            if DEBUG > 1:
                print(f"\nNeighs of point {x} in radius {assignment_radius:g}:")
                for ix in ixs:
                    print(f"At {ix}:", yCoords[ix])
        else:
            # Calculate dists for each target peak if 
            # target is no grid spectrum
            ixs = np.arange(ny, dtype=int)
        
        a = np.zeros(len(ixs))
        if len(ixs) > 0:
            for k in range(ndim):
                a += np.power((yCoords[ixs, k] - x[k]) / scales[k], 2)
        a = np.sqrt(a)
        ix = np.where(a <= assignment_radius)[0]
        distList.append(a[ix])
        indexList.append(ixs[ix])
        
    if DEBUG:
        print("len(distList):", sum(len(l) for l in distList))
    return distList, indexList


def makeMCFNet(
    dist_info,
    weightsY,
    compound_ix,
    weightsX,
    absorption_cost,
    isolated_fit=False,
    ixYinY=None,
    ixXinY=None,
):
    """
    Construct the flow network to route a unit mass through compound nodes C to peak layer X,
    respecting given proportionality constraints, and further to Y with minimal costs.

    dist_info:
        A tuple consisting of a list of distances of X- and Y-nodes
        (i-th element==entries for Y-nodes paired with the i-th X-node)
        and the corresponding indices. Determines the connectivity of the network.
    weightsY:
        capacities of nodes in the target layer
    compound_ix:
        indices of the different compounds in weightsX
    weightsX:
        a list of node weights for the compounds (specified by compound_ix)
    absorption_cost:
        specific cost of absorbed flow
    isolated_fit:
        whether all compounds should be fitted independently. Otherwise a joint fit
        is executed, which avoids overassignment of flow to target nodes.
    ixYinY, ixXinY:
        Boolean arrays, which may exclude indices from matching (mainly used for
        disregarding compound peaks outside the target area, but could also be used
        if match is performed on a restricted range)

    The resulting network has a node indexing which assigns nodes as follows:
    X-nodes: ixs in [0,...,nx-1] with capacities
    Y-nodes (sinks): ixs in [nx,...,nx+ny-1]
    Compound nodes: ix = nx+ny+i, i=0,...,nCompounds-1
    absorbing node (sink): ix = nx+ny+nCompounds
    if not isolated_fit:
        Source node: ix = nx+ny+nCompounds+1
    else:
        Compound nodes are sources

    The layered structure is
    source -> compounds -> X-peaks -> Y-peaks
    """
    assert min(weightsY) >= 0
    assert min(weightsX) > 0

    if ixYinY is None:
        ixYinY = np.arange(len(weightsY))
    if ixXinY is None:
        ixXinY = np.arange(len(weightsX))
    weightsX[~ixXinY] = 0.0
    weightsY[~ixYinY] = 0.0

    compoundIDs = sorted(compound_ix.keys())
    source_flow = sum(weightsY[ixYinY])

    dist_list, index_list = dist_info

    nX, nY, nC = (
        sum([len(ix) for c, ix in compound_ix.items()]),
        len(weightsY),
        len(compound_ix),
    )
    # Y-node and compound-node indexing
    yix = lambda i: nX + i
    cix = lambda i: nX + nY + i
    # Absorbing, ignoring and source nodes
    aix = nX + nY + nC
    # global source (only needed if not isolatedCompoundFit)
    six = nX + nY + nC + 1

    # For convenience, make a dict containing for each compound
    # the node indices and the proportions of its peaks
    compounds = dict()
    for c, ix in compound_ix.items():
        inY = [ixXinY[j] for j in ix]
        if not np.any(inY):
            print(f"No peak of {c} in considered region.")
            compounds[c] = dict(ix=[], p=[], total_weight=0.0)
            continue
        total_weight = sum(weightsX[ix[inY]])
        compounds[c] = dict(
            ix=ix[inY], p=weightsX[ix[inY]] / total_weight, total_weight=total_weight
        )

    # Make the flownet digraph
    G = nx.DiGraph(nX=nX, nY=nY, compounds=compounds, isolated_fit=isolated_fit)
    emptyCompounds = []
    xnodes, cnodes = [], []
    for i, c in enumerate(compoundIDs):
        cInfo = compounds[c]
        if sum(cInfo["p"]) == 0:
            emptyCompounds.append(i)
            continue
        # Weights on X-nodes are proportions wrt compound for X-peaks
        xnodes.extend(
            [(n, {"w": w}) for n, w in zip(cInfo["ix"], cInfo["p"]) if ixXinY[n]]
        )
        if isolated_fit:
            w = source_flow
        else:
            w = 0.0
        cnodes.append((cix(i), {"w": w}))

    G.add_nodes_from(cnodes)
    G.add_nodes_from(xnodes)

    # Ensure that absorbing nodes has sufficient capacity to
    # take up all flow from sources if necessary
    total_source = (
        (nC - len(emptyCompounds)) * source_flow if isolated_fit else source_flow
    )
    G.add_node(aix, w=total_source + 1)

    if not isolated_fit:
        G.add_node(six, w=source_flow)
        # Construct edges source->compounds
        s2cEdges = [
            (six, cix(i)) for i, c in enumerate(compoundIDs) if not c in emptyCompounds
        ]
        G.add_edges_from(s2cEdges)

    # Register X-peaks without Y-neighs
    noNeighs = []
    totalTargetWeight = {}

    for i, c in enumerate(compoundIDs):
        if c in emptyCompounds:
            continue
        cInfo = compounds[c]
        # Construct edges compound -> X-peaks
        c2XEdges = [(cix(i), j) for j, w in zip(cInfo["ix"], cInfo["p"]) if ixXinY[j]]
        G.add_edges_from(c2XEdges)

        if DEBUG > 0:
            totalTargetWeight[c] = 0.0

        for j in cInfo["ix"]:
            if j not in G.nodes:
                noNeighs.append(j)
                continue
            # Construct edges X-peaks -> Y-peaks
            dists, ixs = dist_list[j], index_list[j]
            X2YEdges = [
                (j, yix(ix), w) for ix, w in zip(ixs, dists) if ixXinY[j] and ixYinY[ix]
            ]
            if X2YEdges:
                G.add_weighted_edges_from(X2YEdges)
                if DEBUG > 0:
                    targetWeight_j = sum([weightsY[ixy - nX] for _, ixy, _ in X2YEdges])
                    totalTargetWeight[c] += targetWeight_j
            else:
                if DEBUG > 1:
                    print(
                        "No Y-nodes in assignment neighborhood for X-node %d (x in Y: %s)"
                        % (j, ixXinY[j])
                    )
                noNeighs.append(j)
        if DEBUG > 1:
            print(
                f"\nTotal connected target weight for '{c}': %g" % totalTargetWeight[c]
            )

    # Add weight to ynodes
    nrY = 0
    for n, w in enumerate(weightsY):
        if yix(n) in G.nodes:
            G.nodes[yix(n)]["w"] = w
            nrY += 1
    # Debug
    if DEBUG > 0:
        print("\nTotal connected target nodes: %d" % nrY)
        print("Total target weight: %g\n" % sum(weightsY))

    # Add edge from sources to absorbing node
    if isolated_fit:
        absorptionEdges = [
            (cix(i), aix, absorption_cost)
            for i, c in enumerate(compoundIDs)
            if not c in emptyCompounds
        ]
    else:
        absorptionEdges = [(six, aix, absorption_cost)]
    G.add_weighted_edges_from(absorptionEdges)

    if DEBUG > 1:
        for i in range(nX):
            if i in G.nodes:
                print("       Node %d: %s" % (i, G.nodes[i]))
                print("         preds: %s" % (list(G.predecessors(i))))
                print("         succs: %s" % (list(G.successors(i))))
                if len(list(G.successors(i))) == 0:
                    print("No successors for node %d" % i)
                print(
                    "  w/capacities: %s"
                    % (["%.5f" % G.nodes[n]["w"] for n in G.successors(i)])
                )
            else:
                print("Node %d not in G" % (i))
        if not isolated_fit:
            print("Source node %d:" % six)
            print("   succs: %s" % (list(G.successors(six))))
        print("emptyCompounds: %s" % str(emptyCompounds))
        print("Number of isolated X-nodes: %d" % len(noNeighs))
    return G


def calculateMCF(flownet):
    """
    Solve the linear program, which gives the min cost flow defined by the
    given network with compound nodes
    """
    isolated_fit = flownet.graph["isolated_fit"]
    compounds = flownet.graph["compounds"]
    nX = flownet.graph["nX"]
    nY = flownet.graph["nY"]
    nC = len(compounds)
    compoundIDs = sorted(compounds.keys())

    # Absorbing, ignoring and source nodes
    # Y-node and compound-node indexing
    yix = lambda i: nX + i
    cix = lambda i: nX + nY + i
    aix = nX + nY + nC
    if not isolated_fit:
        six = nX + nY + nC + 1

    # Ordered edges:
    edges = list(flownet.edges())
    nE = len(edges)
    # Reverse map edge->index
    edgeIndices = {e: i for i, e in enumerate(edges)}

    print("Constructing A_eq...")
    tic = default_timer()
    if isolated_fit:
        nEq = 2 * nX + nC
    else:
        nEq = 2 * nX + nC + 1
    A_eq = lil_array((nEq, nE))
    b_eq = np.zeros(nEq)

    # (1) nX+nC(+1) feed-forward flow equalities:
    #     For each node from {X-peaks, compounds(, source)},
    #     the flow sum over all outgoing edges equals
    #     the production plus the incoming flow
    nEq = 0  # reset and use as counting index
    # x-nodes
    deadendC = set()

    for ix, c in enumerate(compoundIDs):
        cInfo = compounds[c]
        peakIndices = cInfo["ix"]

        connected = np.all(
            [
                len(list(flownet.successors(xix))) > 0
                for xix in peakIndices
                if xix in flownet.nodes
            ]
        ) and np.any([j in flownet.nodes for j in peakIndices])
        
        if not connected:
            # no successors → no inflow from compound node
            # Will be treated below adding a zero flow condition
            #    f(s→c) = 0
            # or a full absorption condition
            #    f(c→a) = weight(c)
            # for isolated fitting
            deadendC.add(ix)
            if DEBUG > 1:
                print(
                    f"Encountered dead end(s) at X layer, no assignment for compound {c}"
                )
            continue
        # For all peaks of c, ensure inflow==outflow
        for xix in peakIndices:
            if not xix in flownet.nodes:
                continue
            outEdgeIndices = [edgeIndices[(xix, s)] for s in flownet.successors(xix)]
            inEdgeIndices = [edgeIndices[(p, xix)] for p in flownet.predecessors(xix)]
            A_eq[nEq, inEdgeIndices] = -1.0
            A_eq[nEq, outEdgeIndices] = 1.0
            nEq += 1

    if DEBUG > 0:
        print("Number of dead end compounds: %d / %d" % (len(deadendC), nC))

    # Compound-nodes
    
    # NOTE: This ensures that the total outflow from c-node equals either
    #       its source production (isolated fit) or its inflow.
    #            sum[f(c→x_i)] = f(s→c) (or w_c).
    #       As the next constraints are
    #            f(c→x_i) = p_i*f(s→c) (or p_i*w_c),
    #       the above is implied by sum(p_i)=1 and hence redundant *if* c is 
    #       not empty, i.e. has associated peaks in the considered regions.
    #       if it is empty, we need to add f(s→c)=0 explicitly here. 
    
    if not isolated_fit:
        for i, c in enumerate(compoundIDs):
            cInfo = compounds[c]
            peakIndices = cInfo["ix"]
            if i not in deadendC:
                # Non-empty compound is captured by proportionality constraints below 
                continue
            ix = cix(i)
            if not ix in flownet.nodes:
                continue
            outEdgeIndices = [edgeIndices[(ix, s)] for s in flownet.successors(ix)]
            A_eq[nEq, outEdgeIndices] = 1.0
            inEdgeIndices = [edgeIndices[(p, ix)] for p in flownet.predecessors(ix)]
            A_eq[nEq, inEdgeIndices] = -1.0
            assert list(flownet.predecessors(ix))[0] == six
            nEq += 1
        else:
            # No need to specify outflow c == weight(c),
            # because this follows from the compoundwise
            # specifications, below (proportionality constraints)
            pass

    # Debug
    if DEBUG > 1:
        AA = A_eq.todense()[:nEq, :]
        rk = np.linalg.matrix_rank(AA)
        print("2: rank:%d, nEq:%d" % (rk, nEq))

    # if not isolated_fit (joint compound fit): add global source in last row of A_eq, below

    # (2) nX equalities: flow through compound nodes is proportionally directed to
    #                    the different peaks.
    #                    In case of a global source s, we have
    #                        f(c→xi) = pi*f(s→c) = pi*f(c→X)
    #                    If each c-node is a source with production p:
    #                        f(c→xi) = pi*(p - f(c→a)) = pi*f(c→X)
    #                    <=> f(c→xi) + pi*f(c→a) = pi*p
    for ix, c in enumerate(compoundIDs):
        cInfo = compounds[c]
        peakIndices = cInfo["ix"]
        c2xEdges = [(cix(ix), j) for j in peakIndices if j in flownet.nodes]
        c2xEdgeIndices = [edgeIndices[e] for e in c2xEdges]
        totalWeightC = sum(
            [p for j, p in zip(cInfo["ix"], cInfo["p"]) if j in flownet.nodes]
        )
        if totalWeightC == 0:
            continue
        if isolated_fit:
            c2aEdgeIx = edgeIndices[(cix(ix), aix)]
        else:
            s2cEdgeIx = edgeIndices[(six, cix(ix))]

        if ix in deadendC:
            # There are peaks without successors → no flow to any peak of c
            if isolated_fit:
                # All weight will be absorbed
                A_eq[nEq, c2aEdgeIx] = 1
                b_eq[nEq] = flownet.nodes[cix(ix)]["w"]
            else:
                # Inflow to c must be zero
                A_eq[nEq, s2cEdgeIx] = 1
            nEq += 1
            continue

        proportions = [cInfo["p"][i] / totalWeightC for i in range(len(peakIndices))]
        for p, eix in zip(proportions, c2xEdgeIndices):
            A_eq[nEq, eix] = 1
            if isolated_fit:
                b_eq[nEq] = p * flownet.nodes[cix(ix)]["w"]
                A_eq[nEq, c2aEdgeIx] = p
            else:
                A_eq[nEq, s2cEdgeIx] = -p
            nEq += 1

    # Debug
    if DEBUG > 1:
        AA = A_eq.todense()[:nEq, :]
        rk = np.linalg.matrix_rank(AA)
        print("3: rank:%d, nEq:%d" % (rk, nEq))

    # Add global source node
    if not isolated_fit:
        outEdgeIndices = [edgeIndices[(six, s)] for s in flownet.successors(six)]
        A_eq[nEq, outEdgeIndices] = 1.0
        b_eq[nEq] = flownet.nodes[six]["w"]
        nEq += 1

    # Debug
    if DEBUG > 0:
        cropCheckVal = np.sum(np.abs(A_eq[nEq:, :]))
        assert cropCheckVal == 0

    # Crop unneeded rows
    A_eq = A_eq[:nEq, :]
    b_eq = b_eq[:nEq]
    toc = default_timer()
    print("   took %g secs" % (toc - tic))

    # Debug
    if DEBUG > 1:
        AA = A_eq.todense()[:nEq, :]
        rk = np.linalg.matrix_rank(AA)
        print("4: rank:%d, nEq:%d" % (rk, nEq))

    # (3) Inequalities to account for the maximally matched target weights:
    print("Constructing A_ub...")
    tic = default_timer()
    if isolated_fit:
        # For isolated compound fit: Inflow to ynode from each compound is less than its capacity
        # In order to estimate the number of inequalities, we count
        # for each compound the number of connected Y-nodes
        connectedYNodes = {}
        for c in compoundIDs:
            cInfo = compounds[c]
            peakIndices = cInfo["ix"]
            connectedYNodes[c] = set()
            for xix in peakIndices:
                if xix not in flownet.nodes:
                    continue
                connectedYNodes[c].update(set(flownet.successors(xix)))
            print(
                "Compound %s has %d connected target nodes."
                % (c, len(connectedYNodes[c]))
            )
        nUb = sum([len(ns) for ns in connectedYNodes.values()])
        print("Total estimated number of inequalities is %d." % nUb)
        A_ub = lil_array((nUb, nE))
        b_ub = np.zeros(nUb)

        nUb = 0  # reset to determine the true number of inequalities
        for c in compoundIDs:
            cInfo = compounds[c]
            peakIndices = set(cInfo["ix"])
            for nj in connectedYNodes[c]:
                if DEBUG > 1:
                    if "w" not in flownet.nodes[nj]:
                        print("No weight for Y-Node %d!?" % (nj))
                b_ub[nUb] = max(0.0, flownet.nodes[nj]["w"])
                preds_c = set(flownet.predecessors(nj)).intersection(peakIndices)
                inEdgeIndices = [edgeIndices[(p, nj)] for p in preds_c]
                A_ub[nUb, inEdgeIndices] = 1.0
                nUb += 1
    else:
        # Joint compound fit: Inflow at each ynode is less than its capacity
        # Determine y-nodes with predecessors, because this is the number of
        # inequalities needed
        connectedYNodes = [
            yix(i)
            for i in range(nY)
            if yix(i) in flownet.nodes and len(flownet.pred[yix(i)]) > 0
        ]
        nUb = len(connectedYNodes)
        print(
            "Estimated number of inequalities: %d (nr of connected target nodes)." % nUb
        )
        A_ub = lil_array((nUb, nE))
        b_ub = np.zeros(nUb)
        nUb = 0  # reset to determine the true number of inequalities
        for nj in connectedYNodes:
            if DEBUG > 1:
                if "w" not in flownet.nodes[nj]:
                    print("No weight for Y-Node %d!?" % (nj))
            b_ub[nUb] = max(0.0, flownet.nodes[nj]["w"])
            inEdgeIndices = [edgeIndices[(p, nj)] for p in flownet.predecessors(nj)]
            A_ub[nUb, inEdgeIndices] = 1.0
            nUb += 1
    # regularRows = np.sum(A_ub,1) != 0

    if DEBUG > 0:
        print("True number of inequalities is %d." % nUb)
        cropCheckVal = np.sum(np.abs(A_ub[nUb:, :]))
        assert cropCheckVal == 0

    # Crop unneeded rows
    A_ub = A_ub[:nUb, :]
    b_ub = b_ub[:nUb]
    toc = default_timer()
    print("   took %g secs" % (toc - tic))

    # Check for

    # (3) Costvector
    edgeCosts = np.array([flownet.edges[e].get("weight", 0.0) for e in edges])
    inf_ix = edgeCosts == np.inf
    edgeCosts[inf_ix] = np.max(edgeCosts[np.logical_not(inf_ix)]) + 1000

    print("Running linprog()...")
    tic = default_timer()

    res = linprog(edgeCosts, A_ub, b_ub, A_eq, b_eq, options=dict(disp=True))

    toc = default_timer()
    print("   took %g secs" % (toc - tic))
    if DEBUG > 1:
        print("\nlinprog result:\n", res)
        print("\nlinprog.message:\n%s" % res.message)
    else:
        print("\nlinprog:\n%s" % res.message)

    optFlow = res.x
    for e, f in zip(edges, optFlow):
        flownet.edges[e]["flow"] = f


def flowstats(flownet, includeYStats=False):
    nX, nY = flownet.graph["nX"], flownet.graph["nY"]
    compounds = flownet.graph["compounds"]
    isolated_fit = flownet.graph["isolated_fit"]
    compoundIDs = sorted(compounds.keys())
    nC = len(compoundIDs)

    # Y-node and compound-node indexing
    yix = lambda i: nX + i
    cix = lambda i: nX + nY + i

    # Absorbing nodes
    aix = nX + nY + nC
    if not isolated_fit:
        six = nX + nY + nC + 1

    flowsC2Y = np.zeros(nC)
    costsC2Y, specificCostsC = np.zeros(nC), np.zeros(nC)
    nPeaksConsidered = np.zeros(nC)  # nr of peaks considered for the compound
    inflowsC = np.zeros(nC)
    flowsX2Y = np.zeros(nX)

    for i, c in enumerate(compoundIDs):
        if cix(i) in flownet:
            nPeaksInNet = len(list(flownet.successors(cix(i))))
        else:
            nPeaksInNet = 0
        nPeaksConsidered[i] = nPeaksInNet
        if DEBUG > 1:
            print("\n#Compound '%s'" % c)
        cInfo = compounds[c]
        flowsXC2Y, costsX2Y = [], []
        for n in cInfo["ix"]:
            if n not in flownet.nodes:
                continue
            succs = list(flownet.successors(n))
            flow_n = sum(
                [flownet.edges[(n, s)]["flow"] for s in succs if s not in {aix}]
            )
            flowsXC2Y.append(flow_n)
            flowsX2Y[n] = flow_n
            cost_n = sum(
                [
                    flownet.edges[(n, s)]["flow"] * flownet.edges[(n, s)]["weight"]
                    for s in succs
                    if s not in {aix}
                ]
            )
            costsX2Y.append(cost_n)
            specCost = cost_n / (flow_n) if flow_n != 0 else 0.0
            if DEBUG > 1:
                print(
                    "   Assigned flow from x-node %d (of compound '%s'): %g (specific cost: %g)"
                    % (n, c, flow_n, specCost)
                )
        flowsC2Y[i] = sum(flowsXC2Y)
        costsC2Y[i] = sum(costsX2Y)
        if cix(i) in flownet and not isolated_fit:
            inflowsC[i] = flownet.edges[(six, cix(i))]["flow"]
        else:
            inflowsC[i] = 0.0
        specificCostsC[i] = costsC2Y[i] / flowsC2Y[i] if flowsC2Y[i] != 0 else 0.0

    if isolated_fit:
        # Sources are compound nodes
        absorbedFlow = [
            flownet.edges[(cix(i), aix)]["flow"] if cix(i) in flownet else 0.0
            for i in range(nC)
        ]
        absorbedFlowTotal = sum(absorbedFlow)
    else:
        absorbedFlow = flownet.edges[(six, aix)]["flow"]
        absorbedFlowTotal = absorbedFlow

    if DEBUG > 0:
        print("Total absorbed flow: %g" % absorbedFlowTotal)
    if DEBUG > 1:
        for i, c in enumerate(compoundIDs):
            print("\nCompound %d: '%s'" % (i, c))
            print("  Total compound flow assigned: %g" % flowsC2Y[i])
            if isolated_fit:
                print("  Total absorption: %g" % (absorbedFlow[i]))
            print("  Total compound costs: %g" % costsC2Y[i])
            print("  Specific compound costs: %g" % specificCostsC[i])

    if includeYStats:
        inflowsY = np.zeros(nY)
        for n in range(nY):
            nj = yix(n)
            if nj not in flownet.nodes:
                continue
            preds = flownet.predecessors(nj)
            inflowsY[n] = sum([flownet.edges[(p, nj)]["flow"] for p in preds])
            if nY <= 20:
                print(
                    "Flow/capacity y-node %d: %g/%g"
                    % (n, inflowsY[n], flownet.nodes[nj]["w"])
                )
        residualsY = np.zeros(nY)
        for n in range(nY):
            nj = yix(n)
            v = flownet.nodes.get(nj, None)
            if v:
                residualsY[n] = v["w"] - inflowsY[n]
    else:
        inflowsY, residualsY = None, None

    stats = {
        "inflowsY": inflowsY,
        "inflowsC": inflowsC,
        "residualsY": residualsY,
        "flowsC2Y": flowsC2Y,
        "residualsX": None,
        "costsC2Y": costsC2Y,
        "absorbedFlow": absorbedFlow,
        "specificCostsC": specificCostsC,
        "nPeaksConsidered": nPeaksConsidered,
    }

    return stats


def mcf(
    target_spectrum,
    library,
    assignment_radius=np.inf,
    absorption_cost=None,
    dist_pars={},
    target_id=None,
    lib_id=None,
    savefn=None,
    target_regions=None,
    isolated_fit=False,
    resolveYinResult=False,
    load=True,
    load_dists=True,
):
    """
    Minimum cost flow assignment
    target_spectrum:
        structure with attribute 'coords' and 'weights' describing
        the target spectrum as a list of weighted points
    library:
        dict mapping compound IDs to compound spectra (provided in
        the same form as target_spectrum)
    assignment_radius:
        maximal distance between two peaks for matching them
    absorption_cost:
        specific cost of absorbed flow, default=config.MAX_ABSORPTON_COST
    dist_pars:
        parameter dict for the distance to be used.
        Currently, the only implemented distance is euclidean (2-norm)
        with weighted x- and y-directions. These weights are controlled
        by 'scalex' and 'scaley' with resulting distance

            dist(x,y) = sqrt((x[0]-y[0])**2/scalex**2 + (x[1]-y[1])**2/scaley**2)

        per default scalex=10.0 and scaley=1.0, reflecting scalings for
        C- and H-direction in HSQC spectra.
    target_id:
        allows to specify an ID for the target spectrum. If target_id and lib_id
        are given, the result is saved automatically under

            OUTDIR/<target_id>_<lib_id>_ar<assignment_radius>.pickle

    savefn:
        Allows to provide a specific filepath to save the result. (expects a pathlib.Path)
    isolated_fit:
        whether all compounds should be fitted independently. Otherwise a joint fit
        is executed, which avoids overassignment of flow to target nodes.
    resolveYinResult:
        whether to include inflows and residuals for nodes in target spectrum
        (needs a considerable amount of memory for raster targets, but is required
        for incremental fitting, default: False)
    load:
        If true, the cached result is returned if it exists.
    load_dists:
        If true, the cached dist-lists are used for the weights of peak node connections.
    """
    cache_dir = OUTDIR / "cache" / "mcf"
    if not cache_dir.exists():
        os.makedirs(cache_dir)
        print(f"Created directory '{cache_dir}'")
    if savefn is None and target_id is not None and lib_id is not None:
        savefn = cache_dir / (
            f"mcfResult_{target_id}_by_{lib_id}_ar%g.pickle" % assignment_radius
        )
    if load and savefn is not None and savefn.exists():
        # Load result
        with open(savefn, "rb") as f:
            res = pickle.load(f)
        print(f"Loaded MCF result from '{savefn}'")
        return res

    if absorption_cost is None:
        absorption_cost = MAX_ABSORPTON_COST

    _, weightsY, coordsY = serialize(dict(target_spectrum=target_spectrum))

    compound_ix, weightsX, coordsX = serialize(library)
    target_regions = load_regions(target_regions)

    scalex = dist_pars.get("scalex", DEFAULT_SCALE_C)
    scaley = dist_pars.get("scaley", DEFAULT_SCALE_H)
    scales = np.array([scalex, scaley])

    # Collect timing infos
    timings = {}

    tic = default_timer()
    print("Making dist list for ar=%g ..." % assignment_radius)
    distlist_fn = cache_dir / (
        f"dist_list_{target_id}_{lib_id}_ar%g.pickle" % assignment_radius
    )
    if load_dists and distlist_fn.exists():
        with open(distlist_fn, "rb") as f:
            dist_info = pickle.load(f)
        print(f"Loaded distlists from '{distlist_fn}'")
        computed_dists = False
    else:
        dist_info = euclidean_dists(coordsX, coordsY, scales, assignment_radius)
        with open(distlist_fn, "wb") as f:
            pickle.dump(dist_info, f)
        print(f"Saved distlists to '{distlist_fn}'")
        computed_dists = True
    timings["distList"] = default_timer() - tic
    print("makeDistList() took %g secs" % timings["distList"])

    # Determine which points to exclude based on regions
    print("Checking containment...")
    tic = default_timer()

    ixXinY = getIndicesWithin(
        target_regions,
        coordsX,
        target_id=f"lib {lib_id} in {target_id}",
        cache_dir=cache_dir,
        load=load_dists and not (computed_dists),
    )
    ixYinY = getIndicesWithin(
        target_regions,
        coordsY,
        target_id,
        cache_dir=cache_dir,
        load=load_dists and not (computed_dists),
    )
    timings["containmentCheck"] = default_timer() - tic
    print("Containment checking took %g secs" % timings["containmentCheck"])

    # Generate flow graph
    print("Generating flow net...")
    tic = default_timer()
    flownet = makeMCFNet(
        dist_info,
        weightsY,
        compound_ix,
        weightsX,
        absorption_cost,
        isolated_fit=isolated_fit,
        ixYinY=ixYinY,
        ixXinY=ixXinY,
    )
    toc = default_timer()
    print("Generated flowNet:\n", flownet)
    timings["flowNetGeneration"] = toc - tic
    print("Generation took %g secs" % timings["flowNetGeneration"])

    # Calculate min-cost-flow
    print("Calculating min cost flow...")
    tic = default_timer()
    calculateMCF(flownet)
    toc = default_timer()
    timings["minCostFlow"] = toc - tic
    print("Min-cost-flow calculation took %g secs" % timings["minCostFlow"])

    # Make stats
    print("Collecting stats...")
    tic = default_timer()
    stats = flowstats(flownet, resolveYinResult)
    total_target_weight = sum(weightsY[ixYinY])
    result = MCFResult(
        total_target_weight, stats, absorption_cost, flownet.graph["compounds"], timings
    )
    timings["statsCollection"] = default_timer() - tic
    print("Stats collection took %g secs\n" % timings["statsCollection"])

    if DEBUG > 0:
        print("\nTimings (at nX=%s, nY=%d):" % (len(coordsX), len(weightsY)))
        for k, v in timings.items():
            print("%s: %g secs" % (k, v))

    # Save results
    if savefn is not None:
        with open(savefn, "wb") as f:
            pickle.dump(result, f)
        print(f"Saved MCF result to '{savefn}'")

    return result



import sys
import os
import argparse
import toml
from pathlib import Path
import gzip
import pandas as pd
import numpy as np
import matplotlib as mpl

from mcfnmr.utils.pointspectrum import PointSpectrum
from mcfnmr.core.mcf import mcf
from mcfnmr import __version__
from mcfnmr.config import REPO_URL
from copy import deepcopy
from mcfnmr.routines.utils import (
    singleRun,
    incrementalMCFResultsUpdate,
    updateSavedMCFResult,
)
from mcfnmr.utils.plotting import plot_detected
from pprint import pp


def parse_args():
    arg_parser = argparse.ArgumentParser(
        prog="mcfnmr",
        description=f"""
        MCFNMR  (v{__version__}). A Minimum Cost Flow NMR recombinator.
        """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--config",
        "-c",
        dest="config",
        help=f"Configuration file for the run. See README.md.",
        required=True,
        type=str,
    )
    return arg_parser.parse_args()


def check_config_consistency(config):
    config_fn = config["config_path"]
    config_dir = config_fn.parent
    config.setdefault("isolated_fit", False)
    config.setdefault("incremental_fit", False)
    config.setdefault("output_dir", config_dir / "mcfNMR_output")
    config.setdefault("output_file", None)
    config.setdefault("load", False)
    config.setdefault("plot", False)
    config.setdefault("show_plot", False)
    config.setdefault("gfxformat", "png")

    # Obligatory
    for key in ["lib", "target", "assignment_radius", "detection_threshold"]:
        if key not in config:
            raise Exception(
                f"Config must contain entry '{key}'. Not found in {config_fn}"
            )

    try:
        config["detection_threshold"] = float(config["detection_threshold"])
    except:
        raise Exception(
            f"Couldn't parse float from entry 'detection_threshold' in '{config_fn}': {config['detection_threshold']}"
        )
    try:
        config["output_dir"] = Path(config["output_dir"])
        if not config["output_dir"].is_absolute():
            config["output_dir"] = config_dir / config["output_dir"]
    except:
        raise Exception(
            f"Couldn't parse Path from entry 'output_dir' in '{config_fn}': {config['output_dir']}"
        )

    # Rigid type check
    types = [
        (str, ["lib", "target", "name"]),
        (bool, ["isolated_fit", "incremental_fit", "load"]),
    ]
    for typ, entries in types:
        for e in entries:
            if type(config[e]) is not typ:
                t = type(config[e])
                raise Exception(
                    Exception(
                        f"Entry '{e}' in '{config_fn}' must be of type '{typ}'."
                        f"Found {t}: {config[e]}"
                    )
                )

    for fn in ["target", "lib"]:
        config[fn] = Path(config[fn])
        if not config[fn].is_absolute():
            config[fn] = config_dir / config[fn]

    # Check assignment radius for incremental_fit=True/False
    assignment_radius = config["assignment_radius"]
    if config["incremental_fit"]:
        # assignment_radius should be list of floats
        if type(assignment_radius) is not list:
            raise Exception(
                "For incremental fit, 'assignment_radius' must be a list of floats. "
                f"Found: {assignment_radius} in '{config_fn}'."
            )
        try:
            config["assignment_radius"] = [
                float(r) for r in config["assignment_radius"]
            ]
        except:
            raise Exception(
                "Could not parse floats for all elements of "
                f"'assignment_radius' in '{config_fn}': {assignment_radius}"
            )
    else:
        # assignment_radius should single float
        try:
            config["assignment_radius"] = float(assignment_radius)
        except:
            raise Exception(
                "Could not parse float from "
                f"'assignment_radius' in '{config_fn}': {assignment_radius}"
            )

    supported_gfxformats = mpl.backend_bases.FigureCanvasBase.get_supported_filetypes()
    if config["gfxformat"] not in supported_gfxformats:
        print(
            f"Graphics format '{config['gfxformat']}' is not supported.\nTry one of these:"
        )
        pp(supported_gfxformats)
        raise Exception("Unsupported value for 'gfxformat'.")

    # Hard-coded normalize_assignment (that is, the quantity which is finally compared
    # to the detection_threshold is the assignment to a compound relative to the total
    # weight of the target)
    config["normalize_assignment"] = True


def load_lib(fn):
    fn = Path(fn).absolute()

    if not fn.exists():
        raise Exception(f"File '{fn}' does not exist.")

    gzipped = fn.name[-3:] == ".gz"
    if gzipped:
        libname = ".".join(fn.name.split(".")[:-2])
        f = gzip.open(fn)
    else:
        libname = ".".join(fn.name.split(".")[:-1])
        f = open(fn)
    df = pd.read_csv(f)
    f.close()

    # Treat columns case-insensitive
    df = df.rename(columns={c: c.strip().lower() for c in df.columns})

    required_cols = ["name", "1H", "13C"]
    for c in required_cols:
        if c.lower() not in [col for col in df.columns]:
            raise Exception(
                f"Couldn't find required column '{c}' in library-file '{fn}'."
            )
        elif c != "name":
            df.loc[:, c.lower()] = pd.to_numeric(df.loc[:, c.lower()])

    has_weights = "weight" in df.columns
    if has_weights:
        df.loc[:, "weight"] = pd.to_numeric(df.loc[:, "weight"])

    cpds = df.groupby("name").groups
    print(f"Found %d compounds in library {libname}" % (len(cpds)))

    lib = {}
    for c, ix in cpds.items():
        dfc = df.loc[ix, :]
        if has_weights:
            weights = np.array(dfc.loc[:, "weight"])
        else:
            weights = np.ones(len(dfc))
        coords = np.vstack((dfc["13c"], dfc["1h"])).T
        ptspec = PointSpectrum(coords=coords, weights=weights, name=c)
        lib[c] = ptspec

    return lib, libname


def load_target(fn):
    fn = Path(fn).absolute()

    if not fn.exists():
        raise Exception(f"File '{fn}' does not exist.")

    gzipped = fn.name[-3:] == ".gz"
    if gzipped:
        name = ".".join(fn.name.split(".")[:-2])
        f = gzip.open(fn)
    else:
        name = ".".join(fn.name.split(".")[:-1])
        f = open(fn)
    df = pd.read_csv(f)
    f.close()

    # Treat columns case-insensitive
    df = df.rename(columns={c: c.strip().lower() for c in df.columns})

    required_cols = ["1H", "13C"]
    for c in required_cols:
        if c.lower() not in [col for col in df.columns]:
            raise Exception(
                f"Couldn't find required column '{c}' in target-file '{fn}'."
            )
        else:
            df.loc[:, c.lower()] = pd.to_numeric(df.loc[:, c.lower()])

    if "weight" in df.columns:
        weights = np.array(pd.to_numeric(df.loc[:, "weight"]))
    else:
        weights = np.ones(len(df))
    coords = np.vstack((df["13c"], df["1h"])).T
    ptspec = PointSpectrum(coords=coords, weights=weights, name=name)
    return ptspec, name


def make_point_spectrum(entry, name, ensure_positivity=True):
    if ensure_positivity:
        weights = [max(0.0, float(w)) for w in entry["weights"]]
    else:
        weights = [float(w) for w in entry["weights"]]
    coords = [[float(v) for v in c] for c in entry["coords"]]
    p = PointSpectrum(coords, weights, name=name)
    return p


def make_output_filename(config, outdir, target_name, libname):
    if "name" in config:
        savefn = outdir / ("mcfresult_" + config["name"] + ".pickle")
    else:
        incrstr = "incr" if config["incremental_fit"] else "onepass"
        isostr = "iso" if config["isolated_fit"] else "joint"
        setupstr = f"({isostr}-{incrstr})"
        ar = config["assignment_radius"]
        if config["incremental_fit"]:
            savefn = f"mcfresult_{target_name}_by_{libname}{setupstr}_ar{ar:g}.pickle"
        else:
            # This base name will be used at each incremental step
            savefn = f"mcfresult_{target_name}_by_{libname}{setupstr}.pickle"
    return savefn


def make_mcf_parameters(config, target_name, libname):
    outdir = config["output_dir"]
    if not outdir.is_absolute():
        config_dir = config["config_path"].parent
        outdir = config_dir / outdir
    savefn = make_output_filename(config, outdir, target_name, libname)

    pars = dict(
        assignment_radius=config["assignment_radius"],
        absorption_cost=None,
        dist_pars={},
        target_id=target_name,
        lib_id=libname,
        savefn=savefn,
        target_regions=None,
        isolated_fit=config["isolated_fit"],
        resolveYinResult=False,
        load=config["load"],
        load_dists=config["load"],
    )
    return pars


def classify_result(result, th):
    df = dict(
        compound=[],
        # npeaks=[],
        assigned=[],
        avg_cost=[],
        detection=[],
        original_weight=[],
    )
    compound_ids = sorted(result.compounds)

    for i, cid in enumerate(compound_ids):
        df["compound"].append(cid)
        # df["npeaks"].append(len(result.compounds[cid]["ix"]))
        df["assigned"].append(result.assigned[i])
        df["avg_cost"].append(result.specificCosts[i])
        df["original_weight"].append(result.compounds[cid]["total_weight"])
        df["detection"].append(result.assigned[i] >= th)

    return pd.DataFrame(df)


def save_as_text(df, savefn):
    f = open(savefn, "w")
    df.to_csv(f)
    print(f"Saved results table to '{savefn}'")
    f.close()


def incremental_series(
    target_spectrum,
    library,
    pars,
    load,
    verb=False,
):
    pars = deepcopy(pars)

    pars["absorptionCost"] = None
    pars["specificTargetSpecID"] = target_spectrum.name
    pars["libID"] = f"lib_{len(library)}cpds"
    # Initial setup for AR steps.
    # This triggers incremental fit in singleRun()
    pars["previousARSteps"] = []
    pars["reservedSinkCapacity"] = np.zeros_like(target_spectrum.weights)
    prevResult, prevSavefn = None, None
    results = {}
    for ar in sorted(pars["assignment_radius"]):
        pars["assignmentRadius"] = ar
        prevStepsStr = (
            "[" + ",".join(["%g" % ar for ar in pars["previousARSteps"]]) + "]"
        )
        basefn = ".".join(str(pars["savefn"]).split(".")[:-1])
        savefn = Path(basefn + f"_ar{ar:g}on{prevStepsStr}.pickle")
        result = singleRun(
            pars, library, target_spectrum, load=load, verb=verb, savefn=savefn
        )
        inflowsY = getattr(result, "inflowsY", None)
        loadedResult = inflowsY is None
        if not loadedResult:
            # Update reserved capacity, if not loading results
            incrementalMCFResultsUpdate(result, inflowsY, prevResult)
            # Delete inflowsY and residualsY from result to save storage.
            # We will only need the reserved capacity at the next
            # larger ar-value to extend an incremental series
            result.inflowsY = None
            result.residualsY = None
            updateSavedMCFResult(result, savefn=savefn)
            if prevResult:
                # reservedCapacity not needed for non-maximal ar
                prevResult.reservedSinkCapacity = None
                updateSavedMCFResult(prevResult, savefn=prevSavefn)
        prevSavefn = savefn
        results[ar] = result
        pars["previousARSteps"].append(ar)
        pars["reservedSinkCapacity"] = getattr(result, "reservedSinkCapacity", None)
        prevResult = result
    return results


def load_config(args):
    print(f"\nMCFNMR, version {__version__}")
    config_fn = Path(args.config).absolute()
    config = toml.load(config_fn)
    if getattr(args, "outdir", None) is not None:
        # Only used by test_runner.py, not exposed
        config["output_dir"] = Path(args.outdir).absolute()
        print("\nOption '--outdir' given. Overrides config entry 'output_dir'.")
    config["config_path"] = config_fn
    return config


def run(config):
    mcf_setup = load(config)

    mcf_result = run_mcf(target_spec=mcf_setup["target_spec"], 
            lib=mcf_setup["lib"],
            mcf_pars=mcf_setup["pars"],
            incremental_fit=config["incremental_fit"],
            load=config["load"])
    detection_result = run_detection(mcf_result, 
                                     th=config["detection_threshold"], 
                                     normalize_assignment=config["normalize_assignment"])
    save_as_text(detection_result, config["output_file"])

    if config["plot"]:
        outdir = config["output_file"].parent
        figname = (
            ".".join(config["output_file"].name.split(".")[:-1])
            + "_detections."
            + config["gfxformat"]
        )
        assignment_radius = config["assignment_radius"]
        if config["incremental_fit"]:
            assignment_radius = assignment_radius[-1]
        plot_detected(
            detection_result,
            mcf_setup["lib"],
            mcf_setup["target_spec"],
            assignment_radius,
            outdir / figname,
            mcf_setup["libname"],
            show=config["show"],
        )


def load(config):
    check_config_consistency(config)

    target_spec, target_name = load_target(config["target"])
    nr_neg = np.count_nonzero(target_spec.weights < 0)
    if nr_neg:
        print(f"Setting {nr_neg} negative peak weights to zero.")
        target_spec.weights = np.maximum(target_spec.weights, 0.0)
    lib, libname = load_lib(config["lib"])

    pars = make_mcf_parameters(config, target_name, libname)

    if config["output_file"] is None:
        config["output_file"] = ".".join(pars["savefn"].name.split(".")[:-1] + ["csv"])
        if config["output_dir"] is None:
            config["output_dir"] = pars["savefn"].parent
        config["output_file"] = config["output_dir"] / config["output_file"]

    if not pars["savefn"].parent.exists():
        os.makedirs(pars["savefn"].parent)
        print(f"Created directory '{pars['savefn'].parent}'")
    if not config["output_file"].parent.exists():
        os.makedirs(config["output_file"].parent)
        print(f"Created directory '{config['output_file'].parent}'")
    
    return dict(pars=pars, target_spec=target_spec, lib=lib, libname=libname)


def run_mcf(target_spec, lib, mcf_pars, incremental_fit, load):
    if incremental_fit:
        results = incremental_series(
            target_spectrum=target_spec,
            library=lib,
            pars=mcf_pars,
            load=load,
        )
        result = results[max(results.keys())]
    else:
        result = mcf(
            target_spectrum=target_spec,
            library=lib,
            **mcf_pars,
        )
    return result


def run_detection(result, th, normalize_assignment):
    if normalize_assignment:
        th *= result.originalWeightY
    df = classify_result(result, th)
    return df
    

def main():
    try:
        args = parse_args()
        config = load_config(args)
        run(config)
    except Exception as e:
        print(f"\nmcfNMR failed with error:\n  {str(e)}")
        sys.exit(1)




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.path import Path
from matplotlib.patches import PathPatch
from matplotlib.collections import PatchCollection
import seaborn as sns

from mcfnmr.config import DEFAULT_SCALE_H, DEFAULT_SCALE_C
from mcfnmr.demodata import (
    textbook_abbreviations,
)
from mcfnmr.utils.pointspectrum import (
    makeMix,
    detect_grid,
)
from mcfnmr.routines.utils import figs_dir
from mcfnmr.utils.parsesvg import swapAxes
from mcfnmr.utils.geometry import makeRadiusPoly
from argparse import Namespace
from mcfnmr.demodata.loading import buildHMDBID2NameMap


def sqrt_transfo(max_x):
    def transfo(x):
        res = np.zeros_like(x)
        try:
            iter(x)
            res[x > 0] = np.minimum(np.sqrt(x[x > 0]), max_x)
            res[x < 0] = np.maximum(-np.sqrt(-x[x < 0]), -max_x)
        except:
            res = (
                np.minimum(np.sqrt(x), max_x)
                if x > 0
                else np.maximum(-np.sqrt(-x), -max_x)
            )
        return res

    return transfo


def plotLibOnTarget(
    lib,
    targetSpec,
    libID,
    targetSpecID="",
    assignmentRadius=0.1,
    plotTargetRaster=True,
    densityTransform=None,
    compound_numbers=False,
    plotRadii=True,
    m_alpha=1.0,
    radius_alpha=0.2,
    markerAlphaFromIntensity=True,
    markersize=10,
    scaleSizes=True,
    zoom=None,
    ext="png",
    figsize=None,
    makeLabel=False,
    outdir=figs_dir,
    figtitle=None,
    plotTitle=None,
    show=False,
    ax=None,
):
    if type(lib) is list:
        # Giving a list of libs is used for assigning different colors to peaks (each lib gets a proper one)
        allMixed = []
        for l in lib:
            allMixed.append(makeMix(basis=l, alpha=dict([(k, 1.0) for k in l.keys()])))
            if len(l) == 1:
                allMixed[-1].name = list(l.values())[0].name
    else:
        allMixed = makeMix(basis=lib, alpha=dict([(k, 1.0) for k in lib.keys()]))
    plotPointMixOnRasterTarget(
        allMixed,
        targetSpec,
        targetSpecID,
        libID,
        assignmentRadius=assignmentRadius,
        densityTransform=densityTransform,
        nMix=1,
        outdir=outdir,
        plotTargetRaster=plotTargetRaster,
        flipAxes=True,
        plotRadii=plotRadii,
        m_alpha=m_alpha,
        radius_alpha=radius_alpha,
        makeLabel=makeLabel,
        markersize=markersize,
        compound_numbers=compound_numbers,
        markerAlphaFromIntensity=markerAlphaFromIntensity,
        scaleSizes=scaleSizes,
        plotRegions=False,
        zoom=zoom,
        figsize=figsize,
        ext=ext,
        figtitle=figtitle,
        plotTitle=plotTitle,
        show=show,
        ax=ax,
    )


def plotPointMixOnRasterTarget(
    mix,
    targetSpec,
    targetSpecID,
    libID,
    assignmentRadius,
    nMix,
    outdir,
    flipAxes=True,
    invertAxes=True,
    plotRadii=True,
    plotRegions=True,
    plotTargetRaster=True,
    densityTransform=None,
    figsize=None,
    figsize_zoom=None,
    makeLabel=False,
    markerAlphaFromIntensity=True,
    markersize=10,
    scaleSizes=True,
    compound_numbers=False,
    zoom=None,
    ext="png",
    m_alpha=1.0,
    radius_alpha=0.2,
    show=False,
    figtitle=None,
    plotTitle=None,
    ax=None,
):
    # Don't show if ax was given
    ax_given = ax is not None
    show = show and not ax_given

    if plotTitle is None and plotTargetRaster:
        plotTitle = "targetSpec %s approx. by %s (nMix %s, ar %g)" % (
            targetSpecID,
            libID,
            nMix,
            assignmentRadius,
        )
    elif plotTitle is None:
        plotTitle = "Lib '%s' mix peaks approx %s (nMix %s, ar %g)" % (
            libID,
            targetSpecID,
            nMix,
            assignmentRadius,
        )
    if figtitle is None:
        figfn = outdir / (plotTitle + "." + ext)
    else:
        figfn = outdir / (figtitle + "." + ext)

    flipAxesImPlot = not flipAxes  # imageplot uses different axis ordering

    mpl_style = "default"
    plt.style.use(mpl_style)

    if plotTargetRaster:
        # Raster background plot
        clim = (0, np.max(targetSpec.fullData))
        plot_ranges = {"y": targetSpec.FRanges[0], "x": targetSpec.FRanges[1]}
        ax = plotRasterSpectrum(
            clim=clim,
            title=plotTitle,
            densityTransform=densityTransform,
            spec=targetSpec,
            plotRange=plot_ranges,
            flipAxis=flipAxesImPlot,
            colorbar=False,
            returnAx=True,
            ax=ax,
        )
    else:
        if ax is None:
            _, ax = plt.subplots()
        ax.set_title(plotTitle)
        ax.set_xlabel("")
        if flipAxes:
            ax.set_ylabel("C-shift [ppm]")
            ax.set_xlabel("H-shift [ppm]")
        else:
            ax.set_ylabel("H-shift [ppm]")
            ax.set_xlabel("C-shift [ppm]")

    if plotRegions:
        ax = plotRegionPolys(targetSpec.regions, z0=100, givenAx=ax, flipAxes=flipAxes)

    if plotRadii:
        radii = (assignmentRadius * DEFAULT_SCALE_C, assignmentRadius * DEFAULT_SCALE_H)
    else:
        radii = None

    if type(mix) is list:
        # mix is a list of libs that should be plotted in differen colors
        if targetSpecID == "Ia_01":
            # hack: re-order to control color of specific compounds for plotting OLDB-lib
            #       (this is the only place where I used this approach so far)
            # TODO: rather solve this on caller side!
            # swaps: (21 ,2), (12, 25)
            mix[2], mix[21] = mix[21], mix[2]
            mix[25], mix[12] = mix[12], mix[25]
            print(f"Re-ordering compounds for {targetSpecID}!")

        for i, m in enumerate(mix):
            lab = m.name if makeLabel else None
            lab = textbook_abbreviations.get(lab, lab)
            if makeLabel and compound_numbers:
                lab = ("%d: " % i) + lab

            weights = m.weights if scaleSizes else None
            ax = plotPeaks(
                m.coords,
                weights=weights,
                pcol_ix=i,
                mcol_ix=i,
                radii=radii,
                m_alpha=m_alpha,
                markersize=markersize,
                radius_alpha=radius_alpha,
                iterate_marker_shape=len(mix) > 9,
                alphaFromIntensity=markerAlphaFromIntensity,
                z0=300,
                givenAx=ax,
                flipAxes=flipAxes,
                label=lab,
            )
        if makeLabel:
            # ax.legend(loc=2, ncols=2)
            ax.legend(loc=2, ncols=4)
    else:
        weights = mix.weights if scaleSizes else None
        ax = plotPeaks(
            mix.coords,
            weights=weights,
            radii=radii,
            z0=300,
            m_alpha=m_alpha,
            markersize=markersize,
            radius_alpha=radius_alpha,
            alphaFromIntensity=markerAlphaFromIntensity,
            givenAx=ax,
            flipAxes=flipAxes,
            label=targetSpecID,
        )

    if flipAxes:
        ax.set_xlim(targetSpec.FRanges[1])
        ax.set_ylim(targetSpec.FRanges[0])
    else:
        ax.set_xlim(targetSpec.FRanges[0])
        ax.set_ylim(targetSpec.FRanges[1])
    if invertAxes:
        ax.invert_xaxis()
        ax.invert_yaxis()

    fig = ax.get_figure()
    if figsize_zoom is None:
        figsize = (5, 5)
    fig.set_size_inches(figsize)
    if not ax_given:
        # Caller takes responsibility
        fig.savefig(figfn, dpi=600)
        print("Saved figure '%s'" % figfn)
    if zoom is not None:
        if flipAxes:
            ax.set_xlim(zoom["H"])
            ax.set_ylim(zoom["C"])
        else:
            ax.set_xlim(zoom["C"])
            ax.set_ylim(zoom["H"])
        if invertAxes:
            ax.invert_xaxis()
            ax.invert_yaxis()
        if figtitle is None:
            figzoomfn = outdir / (
                plotTitle + "_zoom%sx%s." % (zoom["H"], zoom["C"]) + ext
            )
        else:
            figzoomfn = outdir / (
                figtitle + "_zoom%sx%s." % (zoom["H"], zoom["C"]) + ext
            )
        ax.legend(loc=2).remove()
        if not ax_given:
            fig.savefig(figzoomfn, dpi=600)
    if show:
        plt.show()
    elif not ax_given:
        plt.close(fig)


def plotRasterSpectrum(
    spec,
    title,
    clim=None,
    plotRange=None,
    flipAxis=False,
    invertAxes=True,
    densityTransform=None,
    colorbar=True,
    cmap="whitejet",
    show=False,
    returnAx=True,
    ax=None,
):
    if cmap == "whitejet":
        # Colormap for spectral intensity
        colors = [
            "white",
            "white",
            "xkcd:red",
            "xkcd:yellow",
            "xkcd:turquoise",
            "xkcd:deep blue",
        ]
        nodes = [0.0, 0.01, 0.15, 0.3, 0.7, 1.0]
        cmap = LinearSegmentedColormap.from_list(
            "intensity_cmap", list(zip(nodes, colors))
        )
    elif cmap in ["grey", "gray"]:
        # Diverging blue->white->red to be transferable back to intensities from image
        colors = [
            "#ffffff",
            "#000000",
        ]
        nodes = [0.0, 1.0]
        cmap = LinearSegmentedColormap.from_list(
            "intensity_cmap", list(zip(nodes, colors))
        )
    elif cmap == "div":
        # Diverging blue->white->red to be transferable back to intensities from image
        colors = [
            "#0000ff",
            "#ffffff",
            "#ff0000",
        ]
        nodes = [0.0, 0.5, 1.0]
        cmap = LinearSegmentedColormap.from_list(
            "intensity_cmap", list(zip(nodes, colors))
        )
    else:
        raise Exception(f"Unknown cmap '{cmap}'")

    if flipAxis:
        data, ranges = spec.fullData.T, (spec.FRanges[1], spec.FRanges[0])
        if plotRange is not None:
            plotRange = {"y": plotRange["x"], "x": plotRange["y"]}
    else:
        data, ranges = spec.fullData, spec.FRanges

    if clim is None:
        clim = (np.min(data), np.max(data))

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    dims = data.shape
    xspan = np.linspace(ranges[0][0], ranges[0][1], dims[0])
    yspan = np.linspace(ranges[1][0], ranges[1][1], dims[1])

    if densityTransform is not None:
        print("max(data) = %g" % np.nanmax(data))
        print("min(data) = %g" % np.nanmin(data))
        data = densityTransform(data)
        print("max(Transform(data)) = %g" % np.nanmax(data))
        print("min(Transform(data)) = %g" % np.nanmin(data))
        clim = (densityTransform(clim[0]), densityTransform(clim[1]))
        m = ax.pcolormesh(yspan, xspan, data, vmin=clim[0], vmax=clim[1], cmap=cmap)
    else:
        print("max(data) = %g" % np.max(data))
        print("min(data) = %g" % np.min(data))
        m = ax.pcolormesh(yspan, xspan, data, vmin=clim[0], vmax=clim[1], cmap=cmap)

    if flipAxis:
        ax.set_ylabel("H-shift [ppm]")
        ax.set_xlabel("C-shift [ppm]")
    else:
        ax.set_ylabel("C-shift [ppm]")
        ax.set_xlabel("H-shift [ppm]")

    if colorbar:
        fig.colorbar(m, ax=ax)

    if plotRange:
        ax.set_xlim(plotRange["x"])
        ax.set_ylim(plotRange["y"])
    ax.set_title(title)

    if invertAxes:
        ax.invert_xaxis()
        ax.invert_yaxis()

    if show:
        plt.show()
    else:
        if returnAx:
            return ax
        else:
            plt.close(fig)


def plotRegionPolys(regions, z0=0, givenAx=None, show=True, flipAxes=False):
    if givenAx is None:
        _, ax = plt.subplots()
    else:
        ax = givenAx

    if regions is None:
        regions = []

    if flipAxes:
        regions = swapAxes(regions)

    for z, poly in enumerate(regions):
        plot_polygon(
            ax, poly, zorder=z0 + z, facecolor="#66666699", edgecolor="#0000FFFF"
        )

    if givenAx is not None:
        return ax
    elif show:
        plt.show()


# Plots a Polygon to pyplot `ax`
def plot_polygon(ax, poly, **kwargs):
    path = Path.make_compound_path(
        Path(np.asarray(poly.exterior.coords)[:, :2]),
        *[Path(np.asarray(ring.coords)[:, :2]) for ring in poly.interiors],
    )

    patch = PathPatch(path, **kwargs)
    collection = PatchCollection([patch], **kwargs)

    ax.add_collection(collection, autolim=True)
    ax.autoscale_view()
    return collection


def plotPeaks(
    coords,
    weights=None,
    pcol_ix=-1,
    mcol_ix=-1,
    markPtIx=None,
    label=None,
    radii=None,
    z0=0,
    m_alpha=1.0,
    alphaFromIntensity=False,
    radius_alpha=0.2,
    markersize=10,
    iterate_marker_shape=True,
    givenAx=None,
    flipAxes=False,
):
    peakColors = sns.color_palette("bright", 9)
    markerColors = peakColors
    markerShapes = [".", "v", "^", "*", "+", "x", "D"]
    if pcol_ix == -1 or mcol_ix == -1:
        pcol = mcol = "black"
        pshp = "."
    else:
        pcol = peakColors[pcol_ix % len(peakColors)]
        mcol = markerColors[mcol_ix % len(markerColors)]
        if iterate_marker_shape:
            pshp = markerShapes[mcol_ix % len(markerShapes)]
        else:
            pshp = markerShapes[0]

    if givenAx is None:
        _, ax = plt.subplots()
    else:
        ax = givenAx

    if weights is None:
        weights = np.ones(len(coords))
    # Normalize weights
    weights = np.array(weights) / sum(weights)
    # srange for weight in [0,1]
    srange = (markersize, markersize)
    # srange = (0.5, 0.5) # For small fig zoom
    sizes = srange[0] + (srange[1] - srange[0]) * weights

    if len(coords) > 0:
        x, y = coords.T
        if radii is not None:
            for xi, yi in zip(x, y):
                if flipAxes:
                    radius = makeRadiusPoly((yi, xi), rx=radii[1], ry=radii[0])
                else:
                    radius = makeRadiusPoly((xi, yi), rx=radii[0], ry=radii[1])
                plot_polygon(ax, radius, color=pcol, alpha=radius_alpha, zorder=z0 + 1)

        if markPtIx is not None:
            if flipAxes:
                ax.scatter(
                    x,
                    y,
                    color="gray",
                    s=sizes,
                    marker=pshp,
                    zorder=z0 + 1,
                    linewidths=0.0,
                )
            else:
                ax.scatter(
                    y,
                    x,
                    color="gray",
                    s=sizes,
                    marker=pshp,
                    zorder=z0 + 1,
                    linewidths=0.0,
                )
            x, y = coords[markPtIx, :].T
            sizes = sizes[markPtIx]
        if alphaFromIntensity:
            alpha = m_alpha + weights * (1.0 - m_alpha) / np.max(weights)
        else:
            alpha = m_alpha
        if flipAxes:
            ax.scatter(
                y,
                x,
                color=mcol,
                s=sizes,
                marker=pshp,
                alpha=alpha,
                zorder=z0 + 3,
                label=label,
            )
        else:
            ax.scatter(
                x,
                y,
                color=mcol,
                s=sizes,
                marker=pshp,
                alpha=alpha,
                zorder=z0 + 3,
                label=label,
            )

    return ax


def plot_detected(df, lib, target, assignment_radius, figname, libname, show=False):
    # Plot point spectrum with detected compounds
    # This is the plotting function called from user interface mcfNMR
    figsize = (10, 8)
    title = ".".join(figname.name.split(".")[:-1])
    show_legend = True

    # Check if the target is generated from a grid spectrum
    # If so, plot target as heatmap.
    res = detect_grid(target.coords)
    grid_spec = res[0] is not None

    predicted = sorted(c for c, d in zip(df["compound"], df["detection"]) if d)

    if grid_spec:
        xspan, yspan, indexer = res
        Z = indexer.makeMatrix(target.weights)
        grid_spec_mock = Namespace(
            FRanges=((min(xspan), max(xspan)), (min(yspan), max(yspan))),
            fullData=Z,
        )
        q95 = np.quantile(target.weights, 0.95)
        ax = plotRasterSpectrum(
            grid_spec_mock,
            title=title,
            colorbar=True,
            returnAx=True,
            densityTransform=sqrt_transfo(q95 * 2),
        )
        fig = ax.get_figure()
    else:
        print(f"Could not deduce grid for spectrum.\n   ({res[1]})")
        fig, ax = plt.subplots(figsize=figsize, layout="constrained")
        FRanges = (
            (min(target.coords[:, 0]), max(target.coords[:, 0])),
            (min(target.coords[:, 1]), max(target.coords[:, 1])),
        )
        grid_spec_mock = Namespace(FRanges=FRanges)

        # Plot target point spectrum
        target_peaks = {target.name: target}
        plotLibOnTarget(
            target_peaks,
            grid_spec_mock,
            libID=libname,
            targetSpecID=target.name,
            assignmentRadius=0.02,
            markersize=0.1,
            radius_alpha=0.2,
            plotTargetRaster=False,
            makeLabel=show_legend,
            ax=ax,
        )

    if libname == "MetaboMiner - Biofluid ( all )":
        ID2name = buildHMDBID2NameMap(verb=0)
        predictions = [
            {ID2name[str(c)]: spec} for c, spec in lib.items() if c in predicted
        ]
    else:
        predictions = [{c: spec} for c, spec in lib.items() if c in predicted]
    title = ".".join(figname.name.split(".")[:-1])
    plotLibOnTarget(
        predictions,
        grid_spec_mock,
        libID=libname,
        targetSpecID=target.name,
        assignmentRadius=assignment_radius,
        plotTargetRaster=False,
        figsize=figsize,
        makeLabel=show_legend,
        plotTitle=title,
        ax=ax,
    )

    fig.set_size_inches(figsize)
    fig.savefig(figname)
    print(f"Saved figure '{figname}'")

    if show:
        plt.show()
    else:
        plt.close(fig)



import os
import urllib.request as rq
from urllib.error import HTTPError
import multiprocessing as mp

import wget
import bs4
import nmrpystar
import numpy as np
import pandas as pd
    
from mcfnmr.config import DATADIR, DEBUG
from pprint import pp
import sys
from argparse import Namespace
import pickle
from mcfnmr.utils.system import get_mcfnmr_home
from mcfnmr.utils.pointspectrum import PointSpectrum
from numpy import vstack
from mcfnmr.utils.spec2csv import saveLibAsCSV

BMRB_DIR = DATADIR / "BMRB"
# This site lists the directory containing .str files for all compounds 
BMRB_FTP_URL = "https://bmrb.io/ftp/pub/bmrb/metabolomics/entry_lists/experimental"

# Destination for str-file download
DOWNLOAD_DIR = BMRB_DIR / "downloads"

HSQC_expnames = [
    # Pre-3.1.1.92
    '2D [1H,13C]-HSQC',
    '2D [1H,13C]-HSQC SW small',
    '2D [1H,13C]-HSQC SW expanded',
    # Post-3.1.1.92
    '2D 1H-13C HSQC',
    '2D 1H-13C HSQC SW small',
    '2D 1H-13C HSQC SW expanded',
    '2D 1H-13C HSQC-TOCSY-ADIA',
    ]
        
known_versions = [
    "NMR STAR v3.1",
    '3.1.0.29',
    "3.1.1.21",
    "3.1.1.31",
    "3.1.1.7",
    "3.1.1.92",
    "3.2.0.16",
    "3.2.1.32",
    "3.2.6.0",
    ]


# HSQC data keys in order of preference
HSQCKeys = [
    "spectral_peak_HSQC", 
    "spectral_peak_1H_13C_HSQC", 
    "spectral_peaks_2D_1H_13C_HSQC_set01",
    "spectral_peaks_2D_1H-13C_HSQC_set01",
    ]

# Keys for the shift table in preference order
# shiftTableKeys = ['Assigned_peak_chem_shift', # This contains multplicities
shiftTableKeys = ['Peak_char',
                  'Spectral_transition_char']
chemShiftField = {'Assigned_peak_chem_shift' : 'Val', 
                  'Peak_char' : 'Chem_shift_val', 
                  'Spectral_transition_char' : 'Chem_shift_val'}
peakIDField = {'Assigned_peak_chem_shift' : 'Peak_ID', 
                  'Peak_char' : 'Peak_ID', 
                  'Spectral_transition_char' : 'Spectral_transition_ID'}

def download_str_file(str_file, dest, download_existent):
        if (dest / str_file).exists() and not download_existent:
            # Skipping download"
            return -1, str_file
        url = "/".join((BMRB_FTP_URL, str_file))
        print(f"Downloading {url} ...")
        try:
            wget.download(url, str(dest / str_file))
            success = True
        except HTTPError as e:
            print(f"Failed to download {str_file}. HTTPError: {e.code}")
            success = False
        return success, str_file


def downloadBMRB(download_existent=False):
    # Download destination for .str files 
    if not DOWNLOAD_DIR.exists():
        os.makedirs(DOWNLOAD_DIR)
        print(f"Created directory {DOWNLOAD_DIR}")
           
    # Use https download
    print(f"Using https connection to download BMRB entries to {DOWNLOAD_DIR} ...")
    
    # Grab site (BMRB_FTP_URL), which lists all relevant .str files
    # and compile a list of their urls
    response = rq.urlopen(BMRB_FTP_URL)
    if response.status != 200:
        print("response.status != 200 (bad url?)")
    if response.msg != "OK":
        print("response.msg != 'OK' (bad url?)")
    doc = "".join(l.decode() for l in response.readlines())
    soup = bs4.BeautifulSoup(doc, 'html.parser')
    str_files = sorted(set(t["href"] for t in soup.find_all("a") if t["href"][-4:]==".str"))
    nstr = len(str_files)   
    print(f"Found {nstr} nmr-star files on BMRB-server.\n\nStarting downloads...\n")
    
    # Download all str-files in parallel
    pool = mp.Pool()
    args = zip(str_files, (DOWNLOAD_DIR for i in range(nstr)), (download_existent for i in range(nstr)))
    results = list(pool.starmap(download_str_file, args))
    failed = [str_file for success, str_file in results if success==False]
    skipped = [str_file for success, str_file in results if success==-1]
    print(f"\nDownloading completed (skipped existing: {len(skipped)}/{nstr}, failed: {len(failed)}/{nstr-len(skipped)}).")


def loopToDataFrame(nmrstar_loop):
    ncols = len(nmrstar_loop.keys)
    coldata = [[r[i] for r in nmrstar_loop.rows] for i in range(ncols)]
    df_exps = dict({c:col for c, col in zip(nmrstar_loop.keys, coldata)})
    return pd.DataFrame(df_exps)


def guessUnknownSpecKey(data):
    candidateHSQCKeys = set()
    # Look for other keys containing HSQC
    for k in data.keys():
        if k.lower().find("hsqc") != -1:
            candidateHSQCKeys.add(k)
    if candidateHSQCKeys:
        print("  Found potential keys: %s"%candidateHSQCKeys)
        print("  Looking for peaklist candidate...")
        good_candis = []
        for k in candidateHSQCKeys:
            candidateSpec = data.get(k, None)
            if candidateSpec is None:
                continue
            hasLoops = getattr(candidateSpec, "loops", None) is not None
            hasShiftTable = False
            if hasLoops:
                for stk in shiftTableKeys:
                    hasShiftTable = hasShiftTable or stk in candidateSpec.loops
            else:
                # Debug
                print("No loops at '%s'"%k)
            goodCandi = hasLoops and hasShiftTable
            if goodCandi:
                good_candis.append(k)
        
        print(f"  Found {len(good_candis)} candidate keys:\n    {good_candis}")
        specKey = good_candis[0]
        print("  I'll try data key '%s'."%k)
    else:
        # Debug
        print("No potential HSQC keys...")
        # print("  Data keys:")
        # pp(list(data.keys()))
        specKey = None
    return specKey


def findSpecKey(data):    
    specKeys = []
    for k in HSQCKeys:
        spec = data.get(k, None)
        if spec is not None:
            specKeys.append(k)
    if specKeys:
        specKey = specKeys[0]
        print("  Using data key '%s'"%k)
    else:
        specKey = None
    return specKey


def reportExperimentList(data):
    if "experiment_list" not in data:
        print("   No entry 'experiment_list' in save frames.")
        HSQC_exp_available = False
        explist = None
    else:
        explist = data["experiment_list"].loops["Experiment"]
        
        df_exps = loopToDataFrame(explist)
        print("Experiment list:")
        pp([n for n in df_exps["Name"]])
        HSQC_exp_available = np.any([n in df_exps["Name"] for n in HSQC_expnames])
        
        # Debug
        if DEBUG:
            newkeys = [n not in HSQC_expnames and n.lower().find("hsqc") != -1 for n in df_exps["Name"]]
            if np.any(newkeys):
                print("print new HSQC key?")
                print(df_exps["Name"][newkeys])
                sys.exit()
            
    # # Assigned_chemical_shifts holds experiments as well
    # chem_shifts = data['assigned_chemical_shifts'].loops
    # exps = loopToDataFrame(chem_shifts["Chem_shift_experiment"])
    
    return HSQC_exp_available


def addBasicEntryInformation(p, entry):
    entry_info = p.value.saves["entry_information"].datums
    entry.name = p.value.name
    entry.title = entry_info["Title"].strip()
    entry.version = entry_info["NMR_STAR_version"]
    entry.orig_version = entry_info.get("Original_NMR_STAR_version", None)


def addCompoundInformation(data, entry):
    chem_comp_1 = data["chem_comp_1"]

    entry.smiles = None
    smilesTable = chem_comp_1.loops.get("Chem_comp_SMILES", None)
    if smilesTable:
        for row in smilesTable.rows:
            if row[0] == "canonical":
                entry.smiles = row[1]
                break
    
    entry.inchi = chem_comp_1.datums.get("InChI_code", None)
    if entry.inchi is not None:
        inchi = entry.inchi.split("=")[1]
        
    entry.CASRegistries = set()
    entry.PubChemCodes = set()
    chem_comp_1_name = chem_comp_1.datums["Name"]
    entry.molNames = {chem_comp_1_name}
    
    linkTable = chem_comp_1.loops.get("Chem_comp_db_link", None)
    if linkTable is None:
        print("  No db link table found.")
    else:
        CASKeys = ["CAS Registry", "CAS"]
        PubChemKeys = ["PubChem"]
        df = loopToDataFrame(linkTable)
        entry.PubChemCodes = list(df.loc[np.isin(df["Database_code"], PubChemKeys), "Accession_code"])
        entry.CASRegistries = list(df.loc[np.isin(df["Database_code"], CASKeys), "Accession_code"])
        entry.molNames = set(df["Entry_mol_name"]) if "Entry_mol_name" in df.columns else set()
            
    print("   Nr of molecule names: %d"%len(entry.molNames))
    print("     names:", entry.molNames)
    if "." in entry.molNames:
        entry.molNames = entry.molNames.difference(".")
        print("      Removing '.' from names...")
    print("   Nr of CAS registries: %d"%len(entry.CASRegistries))
    print("     codes:", entry.CASRegistries)
    print("   Nr of PubChem codes: %d"%len(entry.PubChemCodes))
    print("     codes:", entry.PubChemCodes)
    print("InChI code: %s"%entry.inchi)
    print("SMILES code: %s"%entry.smiles)
    

def parseBMRBEntry(fn):    
    with open(fn, "r") as f:
        nmrStarStr = f.read()
    p = nmrpystar.parse(nmrStarStr)       
    data = p.value.saves
    
    entry = Namespace(peaks={})
    addBasicEntryInformation(p, entry)

    if entry.name is None:
        entry.msg = "Couldn't pars substance for file '%s'"%fn
        print(entry.msg)
        return False, entry
    
    print("parseBMRBEntry()\n  Parsed NMRstar data for compound '%s' ('%s') from '%s'"%(entry.title, entry.name, fn))
    print(f"  NMR-STAR version: {entry.version} (orig: {entry.orig_version})")

    # Check if experiment_list registers a HSQC experiment
    HSQC_exp_available = reportExperimentList(data)
    
    if DEBUG:
        if entry.version not in known_versions:
            print(f"new version '{entry.version}'")
            sys.exit()
            
    
    if "chem_comp_1" not in data:
        entry.msg = f"  Couldn't find compound information for '%s' (NMR-START version {entry.version}) - skipping entry."%entry.name
        print(entry.msg)
        return False, entry
    
    addCompoundInformation(data, entry)
    
    specKey = findSpecKey(data)
    if specKey is None:
        # Try to find an alternative key for the HSQC spectrum
        print(f"  None of the specified HSQCKeys present in data. NMR-STAR version {entry.version}")
        specKey = guessUnknownSpecKey(data)
    
    entry.specKey = specKey    
    spec = None if specKey is None else data[specKey]
    
    if spec is None:
        print(f"  No peak (HSQC) data found for '%s' (NMR-STAR version '{entry.version}')"%entry.name)
        print(f"     (HSQC experiment mentioned in experiment_list? {HSQC_exp_available})")
        # Debug
        if DEBUG:
            if HSQC_exp_available:
                print("HSQC experiment mentioned but not found.")
                sys.exit()
    else:
        shiftTable = None
        for k in shiftTableKeys:
            shiftTable = spec.loops.get(k, None)
            if shiftTable is not None:
                print("  Using shift table key '%s'"%k)
                shiftTableKey = k
                entry.shiftTableKey = k
                break
        if shiftTable is None:
            entry.msg = "Found no shift table for '%s'"%entry.name
            print(entry.msg)
            return True, entry
        for i, _ in enumerate(shiftTable.rows):
            d = shiftTable.getRowAsDict(i)
            pid = d[peakIDField[shiftTableKey]]
            entry.peaks.setdefault(pid, {})
            dim = d["Spectral_dim_ID"]
            entry.peaks[pid].setdefault(dim, [])
            entry.peaks[pid][dim].append(d[chemShiftField[shiftTableKey]])
        print("   Nr of peaks: %d"%len(entry.peaks))
        for i, pid in enumerate(sorted(entry.peaks.keys())):
            print("     peak %d: %s"%(i, entry.peaks[pid]))
    return True, entry


def parseBMRBEntries(parallel=True):
    file_list = list(DOWNLOAD_DIR.iterdir())
    offset = 0
    if parallel:
        pool = mp.Pool()
        results = pool.map(parseBMRBEntry, file_list[offset:])
    else:
        results = map(parseBMRBEntry, file_list[offset:])
    entries = []
    failed = []
    for i, res in enumerate(results):
        print("\nEntry ", i+offset)
        success, entry = res
        if not success:
            print(f"Parsing failed for '{file_list[offset+i]}'.")
            failed.append(file_list[offset+i])
            continue
        pp(entryAsDict(entry))
        entries.append(entry)
    return entries, failed
        
    
def entryAsDict(e):
    attrs = [attr for attr in dir(e) if attr[0]!="_"]
    return {a:getattr(e, a) for a in attrs}


def spectrumFromEntry(e):
    print(f"\nEntry {e.name}")
    pp(e.peaks)
    df = pd.concat([pd.DataFrame(p) for p in e.peaks.values()])
    
    if not np.all(np.isin(["1", "2"], list(df.columns))):
        print(f"peak ids '1' or '2' not found for {e.name}")
        if ("?" in list(df.columns)):
            print(f"Spectrum for {e.name} might be incomplete... skipping.")
            return None
        print(df)
        if len(df.columns) < 2:
            return None
        else:
            sys.exit(1)
    
    coordsH, coordsC = [float(c) for c in df["1"]], [float(c) for c in df["2"]]
    
    # Debug
    if np.mean(coordsH) > np.mean(coordsC):
        print(f"coord order seems to be wrong for {e.name}. Switching '1' for '2'.")
        print(df)
        coordsH, coordsC = coordsC, coordsH
        
    coords = np.vstack((coordsC, coordsH)).T
    ptspec = PointSpectrum(coords, weights=np.ones_like(coordsH), name=e.name)
    return ptspec
    

def makeBMRBLib(entries):
    lib = {}
    nopeaks = []
    for e in entries:
        if e.peaks:
            if e.name in lib:
                print(f"Duplicate entry '{e.name}'!")
                print("Version 1:")
                print(lib[e])
                print("Version 2:")
                pp(entryAsDict(e))
                # Debug
                if DEBUG:
                    sys.exit(1)
            spec = spectrumFromEntry(e)
            if spec is None:
                nopeaks.append(e.name)
            else:
                lib[e.name] = spec
        else:
            nopeaks.append(e.name)
    return lib, nopeaks


def prepareBMRB():
    a = 0
    while a==0:
        a = input("\nPrepare BMRB lib? (This takes a few minutes) (Y/n) ")
        if a.strip() in ["", "Y", "y", "yes"]:
            break
        elif a.strip() in ["n", "N", "no"]:
            return
        else:
            a=0

    if not (BMRB_DIR).exists():
        os.makedirs(BMRB_DIR)
        print(f"Created directory {BMRB_DIR}")

    libfile = BMRB_DIR / "BMRBlib.csv"     
    if libfile.exists():
        a=0
        while a==0:
            a = input(f"File '{libfile}' already exists.\nRe-generate? (Y/n) ")
            if a.strip() in ["", "Y", "y", "yes"]:
                break
            elif a.strip() in ["n", "N", "no"]:
                return
            else:
                a=0
    
    downloadBMRB()
    entries, failed = parseBMRBEntries()
    
    # # Save intermediate results for faster debug
    # entries_fn = get_mcfnmr_home() / "tmp/entriesBMRB.pickle"
    # with open(entries_fn, "wb") as f:
    #     pickle.dump((entries, failed), f)
    # print(f"Saved entries to '{entries_fn}'")
    
    # with open(entries_fn, "rb") as f:
    #     entries, failed = pickle.load(f)
        
    print(f"\nParsed {len(entries)} entries from {len(entries)+len(failed)} files (failed: {len(failed)})")    
    
    lib, nopeaks = makeBMRBLib(entries)
    
    print(f"\nCreated library of {len(lib)} compounds, didn't find peaks for {len(nopeaks)} entries.")
    
    saveLibAsCSV(lib, "BMRBlib", libfile)
    
    print("\nPreparing BMRB done.\n")

if __name__=="__main__":
    prepareBMRB()


import os
import platform
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pickle
import shutil
from pathlib import Path
from argparse import Namespace
import multiprocessing as mp

import wget
import waybackpack as wbp

from mcfnmr.config import (
    UCSFDIR,
    PROCESSED_UCSFDIR,
    METABOMINER_DIR,
    COLMAR_DATADIR,
    MM_PEAKLISTDIR,
)
from mcfnmr.demodata.loading import loadMMLibrary, loadMMMixPointSpectra
from mcfnmr.utils.loading import parseUCSF
from mcfnmr.utils import spec2csv
from mcfnmr.utils.spec2csv import saveLibAsCSV
from mcfnmr.demodata import binning

def waybackDownload(component, fn, url):
    if not fn.exists():
        print(f"Downloading {component} to '{fn}' ...")
        p = wbp.pack.Pack(url)
        p.timestamps = [p.timestamps[-1]]  # only download most recent
        p.download_to(METABOMINER_DIR / f"{component}_tmp")
        path = METABOMINER_DIR / (
            "/".join([f"{component}_tmp", p.timestamps[0]] + url.split("/")[2:])
        )
        shutil.move(path, fn)
        shutil.rmtree(METABOMINER_DIR / f"{component}_tmp")
        print("Done.")
    else:
        print(f"Skipping {component} download. {fn} already exists.")
        
    
def download_metabominer_data(ucsf_url, expls_url, mm_url, sparky_url):
    # Below, we use waybackpack to download MetaboMiner from archive.org, it uses urls like
    # "https://web.archive.org/web/20140502110649/http://wishart.biology.ualberta.ca/metabominer/downloads/..."

    ucsf_fn = METABOMINER_DIR / "ucsf.zip"
    expls_fn = METABOMINER_DIR / "examples.zip"
    mm_fn = METABOMINER_DIR / "metabominer.zip"
    sparky_fn = METABOMINER_DIR / "sparky-linux2.6-64bit.tar.gz"

    if not sparky_fn.exists():
        print(f"Downloading sparky to '{sparky_fn}'.")
        wget.download(sparky_url, str(sparky_fn))
        print("Done.")
        
    args = [
        ("MetaboMiner UCSF files", ucsf_fn, ucsf_url),
        ("MetaboMiner examples", expls_fn, expls_url),
        ("MetaboMiner", mm_fn, mm_url),
        ]
    pool = mp.Pool()
    _ = list(pool.starmap(waybackDownload, args))
    
    print("Unpacking sparky ...")
    shutil.unpack_archive(sparky_fn, METABOMINER_DIR)
    print("Unpacking ucsf files ...")
    shutil.unpack_archive(ucsf_fn, METABOMINER_DIR)
    if expls_fn.exists():
        print("Unpacking example files ...")
        shutil.unpack_archive(expls_fn, METABOMINER_DIR)
        print("Copying DEEP Picker peak lists into peak list directory ...")
        for fn in (COLMAR_DATADIR/"Deep Picker unreferenced for MCF").iterdir():
            if fn.name[-8:] == "hsqc.txt":
                shutil.copy2(fn, MM_PEAKLISTDIR / fn.name)
                print(f"   {fn.name}")
    if mm_fn.exists():
        print("Unpacking MetaboMiner files ...")
        shutil.unpack_archive(mm_fn, METABOMINER_DIR)


def save_MMlib_as_csv():
    name = "Biofluid ( all )"
    lib = loadMMLibrary(libName=name)
    libname = "MetaboMiner - " + name
    saveLibAsCSV(lib, libname, METABOMINER_DIR / "csv" / (libname+".csv"))


def process_ucsf_spectra(load=True, plot=False, show=False):
    if not (PROCESSED_UCSFDIR).exists():
        os.makedirs(PROCESSED_UCSFDIR)
        print(f"Created directory {PROCESSED_UCSFDIR}")

    files = {}
    for fn in os.listdir(UCSFDIR):
        if fn.split("_")[1] == "tocsy.ucsf":
            # Skip tocsy
            continue
        files[fn[:4]] = os.path.join(UCSFDIR, fn)

    for fid, ucsfFile in files.items():
        # Save .pickle files with matrix and header info
        print("Processing '%s'" % fid)
        saveFile = PROCESSED_UCSFDIR / (fid + ".pickle")
        if load and os.path.exists(saveFile):
            with open(saveFile, "rb") as file:
                data = pickle.load(file)
            print("Loaded data '%s'" % saveFile)
            header = data["header"]
            matrix = data["matrix"]
        else:
            matrix, header = parseUCSF(Path(ucsfFile), outdir=PROCESSED_UCSFDIR)
            header["mixID"] = fid

            # Pickling
            with open(saveFile, "wb") as file:
                pickle.dump({"matrix": matrix, "header": header}, file)
            print("Saved data '%s'" % saveFile)

        # Save grid-csvs for user interface testing
        ucsf_file = UCSFDIR / f"{fid}_hsqc.ucsf"
        bindims = binning[fid]
        outfn = METABOMINER_DIR / "csv" / f"{fid}({bindims[0]}x{bindims[1]}).csv.gz"
        if not (outfn.exists() and load):
            args = Namespace(infile=ucsf_file, bindims=bindims)
            pt_spec = spec2csv.ucsf2ptspec(args, name=fid)
            spec2csv.write_ptspec_as_csv(pt_spec, fn=outfn)

        if plot:
            plt.imshow(matrix, origin="lower")
            figname = os.path.join(PROCESSED_UCSFDIR, fid + "_matrix.png")
            plt.savefig(figname)

        if show:
            plt.show()
        else:
            plt.close("all")


def convert_peaklists():
    # Convert MetaboMiner peak lists into csv files with
    # columns 1H, 13C, and weights instead of w1, w2, and Data height
    ptSpectra = loadMMMixPointSpectra()
    for mixID in ptSpectra:
        df = {
            "weights": ptSpectra[mixID].weights,
            "1H": ptSpectra[mixID].coords[:, 1],
            "13C": ptSpectra[mixID].coords[:, 0],
        }
        fn = METABOMINER_DIR / "csv" / f"{mixID}(peaklist).csv"
        pd.DataFrame(df).to_csv(fn, sep=",")
        print(f"Wrote {fn}")


def prepareMM():
    ucsf_url = "http://wishart.biology.ualberta.ca/metabominer/downloads/ucsf.zip"
    expls_url = "http://wishart.biology.ualberta.ca/metabominer/downloads/examples.zip"
    mm_url = "http://wishart.biology.ualberta.ca/metabominer/downloads/metabominer.zip"
    if platform.system() == "Windows":
        sparky_url = "https://www.cgl.ucsf.edu/home/sparky/distrib-3.114/sparky-win32.zip"
        raise Exception("Sparky on Windows not supported yet")
    elif platform.system() == "Linux":
        sparky_url = "https://www.cgl.ucsf.edu/home/sparky/distrib-3.115/sparky-linux2.6-64bit.tar.gz"
    elif platform.system() == "Darwin":
        # MacOS
        sparky_url = "https://www.cgl.ucsf.edu/home/sparky/distrib-3.115/sparky-mac10.4-intel.dmg"
        raise Exception("Sparky on MacOS not supported yet")

    if not (METABOMINER_DIR).exists():
        os.makedirs(METABOMINER_DIR)
        print(f"Created directory {METABOMINER_DIR}")

    download_metabominer_data(ucsf_url, expls_url, mm_url, sparky_url)

    if not (METABOMINER_DIR / "csv").exists():
        os.makedirs(METABOMINER_DIR / "csv")
        print(f"Created directory {METABOMINER_DIR / 'csv'}")

    save_MMlib_as_csv()
    process_ucsf_spectra()
    convert_peaklists()

    print("\nPreparing MetaboMiner data done.\n")


if __name__=="__main__":
    prepareMM()



import numpy as np
from pprint import pprint

from mcfnmr.demodata import case_dict
from mcfnmr.demodata.loading import (
    loadMMLibrary,
    loadMMMixtureLists,
    loadMMMixPointSpectra,
    loadMetaboMinerSpectrum,
)
from mcfnmr.demodata import mixNamesOLDB
from mcfnmr.routines.utils import (
    runMetaboMinerFits,
    prepareParameters,
    cache_dir,
    totalWeights,
    makeResultsDF,
    filterDF,
    collect_and_serialize_MM_grid_spectra,
)
from mcfnmr.utils.pointspectrum import makeRasterPointSpectrum
from mcfnmr.demodata import binning
from mcfnmr.config import DEBUG


def indicateContainment(df, th, containedCompounds, scale_by_peaknr):
    compounds = sorted(set(df["compound"]))
    assert len(compounds) == len(df["compound"])

    results = {}
    for c, assignment, npeaks in zip(df["compound"], df["assigned"], df["nPeaks"]):
        # If more flow than th is assigned to the compound in total, it is indicated as contained
        if th == 0.0:
            predicted = assignment > th
        elif scale_by_peaknr:
            predicted = False if npeaks == 0 else assignment >= th * npeaks
        else:
            predicted = assignment >= th
        # True containment
        contained = c in containedCompounds

        results[c] = (predicted, contained)
    
    for c in containedCompounds:
        if c in results:
            continue
        else:
            # c not contained in lib
            if DEBUG > 1:
                print(f"Compound '{c}' not in library, therefore it is not predicted.")
            results[c] = (False, True)
            pass
            
    return results


def classifyCompounds(
    dfCompounds, pars, th, containedCompounds, scale_by_peaknr, verb=False
):
    libID, targetID = pars["libID"], pars["targetSpecID"]
    targetName = mixNamesOLDB.get(targetID, targetID)

    filters = {"lib": libID, "target": targetName}
    dfC = filterDF(dfCompounds, filters)

    classificationResult = indicateContainment(
        dfC, th, containedCompounds, scale_by_peaknr
    )

    tp = [c for c, r in classificationResult.items() if r == (True, True)]
    tn = [c for c, r in classificationResult.items() if r == (False, False)]
    fp = [c for c, r in classificationResult.items() if r == (True, False)]
    fn = [c for c, r in classificationResult.items() if r == (False, True)]

    if verb:
        print("\nFN['%s'] = %s" % (targetID, str(fn)))
        print("FP['%s'] = %s" % (targetID, str(fp)))

    npos = len(tp) + len(fp)
    if npos:
        precision = len(tp) / npos
    else:
        precision = 0.0
    ntrue = len(tp) + len(fn)
    if ntrue:
        recall = len(tp) / ntrue
    else:
        recall = 1

    fscore = (
        precision * recall * 2 / (precision + recall)
        if (precision + recall) > 0
        else 0.0
    )

    res = dict(
        recall=recall,
        precision=precision,
        fscore=fscore,
        TP=tp,
        FP=fp,
        FN=fn,
        compound_df=dfC,
    )
    return res


def classificationStats(
    data, pars, th, mixSubs, mixIDs, scale_by_peaknr=False, verb=False
):
    stats = {}
    for mixID in mixIDs:
        dfCompounds = data[mixID]["compounds"]
        pars["targetSpecID"] = data[mixID]["targetSpecID"]

        if verb:
            print(f"\nMix '{mixID}'")

        if mixID[-4:] == "(dp)":
            mixID_stripped = mixID[:-4]
        else:
            mixID_stripped = mixID
        contained = mixSubs[mixID_stripped]

        cl = classifyCompounds(
            dfCompounds,
            pars,
            th,
            containedCompounds=contained,
            scale_by_peaknr=scale_by_peaknr,
            verb=verb,
        )

        if verb:
            print("\n%s:" % mixID)
            for k in cl:
                if k in ["FN", "FP", "TP"]:
                    print(f"'{k}': %d" % len(cl[k]))
                elif k == "compound_df":
                    pass
                else:
                    print(f"'{k}': %g" % cl[k])

        stats[mixID] = {"classificationThreshold": th, "classification": cl}

    return stats


def runSingleClassificationIncremental(caseSpec, pars, arSpan, th, load=True):
    # This calculates a series of incremental fits for the ar in arSpan
    # And gives the result for the maximal ar
    assert caseSpec["incrementalFit"]

    mixIDs = caseSpec["mixIDs"]
    mixSubs = loadMMMixtureLists()
    if pars["pointTarget"]:
        mixSpecs = loadMMMixPointSpectra()
    else:
        pars["binning"] = None
        mixSpecs = collect_and_serialize_MM_grid_spectra(mixIDs, pars)
    lib = loadMMLibrary(pars["libID"])

    pars["arSpan"] = arSpan
    results = runMetaboMinerFits(
        mixIDs=mixIDs,
        base_pars=pars,
        lib=lib,
        mixSpecs=mixSpecs,
        # pointTarget=caseSpec["pointTarget"],
        returnAllResults=True,
        load=load,
    )

    dfs = {}
    for mixID in mixIDs:
        result = results[mixID][max(arSpan)]
        dfTargets, dfCompounds = makeResultsDF(
            [{"res": result, "pars": pars}], originalWeights=totalWeights(lib)
        )
        dfs[mixID] = {
            "targets": dfTargets,
            "compounds": dfCompounds,
            "targetSpecID": pars["targetSpecID"],
            "targetTotalWeight": result.originalWeightY,
        }

    stats = {}
    for ar in arSpan:
        pars["assignmentRadius"] = ar
        stats[ar] = classificationStats(dfs, pars, th, mixSubs, mixIDs)

    print("Classification results:")
    pprint(stats[max(arSpan)])

    return stats


def runSingleClassification(caseSpec, pars, ar, th, load=True):
    assert not caseSpec["incrementalFit"]
    mixIDs = caseSpec["mixIDs"]
    mixSubs = loadMMMixtureLists()
    lib = loadMMLibrary(pars["libID"])

    if pars["pointTarget"]:
        mixSpecs = loadMMMixPointSpectra()
    else:
        pars["binning"] = None
        mixSpecs = collect_and_serialize_MM_grid_spectra(mixIDs, pars)

    pars["assignmentRadius"] = ar
    results = runMetaboMinerFits(
        base_pars=pars, lib=lib, mixSpecs=mixSpecs, mixIDs=mixIDs, load=load
    )
    dfs = {}
    for mixID in mixIDs:
        result = results[mixID]
        dfTargets, dfCompounds = makeResultsDF(
            [{"res": result, "pars": pars}], originalWeights=totalWeights(lib)
        )
        dfs[mixID] = {
            "targets": dfTargets,
            "compounds": dfCompounds,
            "targetSpecID": pars["targetSpecID"],
            "targetTotalWeight": result.originalWeightY,
        }
    scores = classificationStats(dfs, pars, th, mixSubs, mixIDs)

    print("Classification results:")
    pprint(scores)

    return scores


def main(cfg):
    assert cfg["task"] == "classification"
    assert cfg["lib"] == "MetaboMiner"
    assert cfg["mix"].split("_")[0] in ["N925", "N926", "N987", "N988"]
    assert cfg["setup"] in ["A", "B", "C", "D", "E", "F"]

    target_info = cfg["mix"].split("_")
    if len(target_info) > 1:
        smoothingRadius = float(target_info[2])
    else:
        smoothingRadius = 0.0
    mixID = target_info[0]

    # assignment radii (hard-coded for incremental approach)
    arSpan = [0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1, 0.12, 0.15, 0.175, 0.2, 0.25]
    arSpan = np.array(sorted(arSpan))

    setupSpec = case_dict[cfg["setup"]][mixID]
    incrementalFit = setupSpec["incrementalFit"]

    kwargs = dict(
        incrementalFit=incrementalFit,
        isolated_fit=setupSpec["isolated_fit"],
        noiseFilterLevel=None,
        pointTarget=setupSpec["pointTarget"],
        smoothing=smoothingRadius,
        workingDir=cfg["outdir"],
        recompute=cfg.get("recompute", True),
    )
    pars = prepareParameters(
        specDomain="metabominer", targetSpecID=cfg["mix"], libID=cfg["lib"], **kwargs
    )

    load = not pars["recompute"]
    if incrementalFit:
        ars = np.append(arSpan[arSpan < cfg["ar"]], cfg["ar"])
        _ = runSingleClassificationIncremental(
            setupSpec, pars, arSpan=ars, th=cfg["th"], load=load
        )
    else:
        _ = runSingleClassification(
            setupSpec, pars, ar=cfg["ar"], th=cfg["th"], load=load
        )


if __name__ == "__main__":
    # Test run
    cfg = dict(
        lib="MetaboMiner",
        recompute=False,
        task="classification",
        setup="A",
        mix="N925",
        # optimal for setup A
        ar=0.08,
        th=0.003,
        # mix="N926",
        # # optimal for N926
        # ar = 0.03,
        # th = 5e-05,
        outdir=cache_dir,
    )

    main(cfg)



import pickle
import os
import numpy as np
from shapely.geometry import Point, Polygon
from shapely.affinity import scale
from mcfnmr.utils import parsesvg


class RectRegion(object):
    def __init__(self, xrange, yrange):
        self.xrange = xrange
        self.yrange = yrange

    def hitTest(self, xy):
        x, y = xy
        return (
            x >= self.xrange[0]
            and x <= self.xrange[1]
            and y >= self.yrange[0]
            and y <= self.yrange[1]
        )


def load_regions(regions):
    if regions:
        if type(regions) is dict:
            res = parsesvg.loadRegions(**regions)
        else:
            # regions already provided as Polygon list
            assert type(regions) is list
            res = regions
    else:
        res = []
    return res


def getIndicesWithin(regions, coords, target_id=None, cache_dir=None, load=False):
    """
    Return coord indices, which are not part of any given region
    """
    if target_id is not None:
        if cache_dir is None:
            raise Exception("Need working directory to use index-lookup-cache.")
        fn = cache_dir / (target_id + "_region_ix.pickle")
        if load and os.path.exists(fn):
            with open(fn, "rb") as f:
                ix = pickle.load(f)
            print("Loaded indices from %s" % fn)
            return ix
    else:
        fn = None

    if len(regions) == 0:
        return np.ones(coords.shape[0], dtype=bool)
    regionUnion = regions[0]
    for i in range(1, len(regions)):
        regionUnion = regionUnion.union(regions[i])
    ix = np.array([regionUnion.contains(Point(xy)) for xy in coords])

    if fn is not None:
        with open(fn, "wb") as f:
            pickle.dump(ix, f)
        print("Saved indices to %s" % fn)

    return ix


def makeRadiusPoly(xy, rx, ry, typ="ellipse"):
    # Make a square
    x, y = xy
    if typ == "rect":
        # draw rect
        path = ((x - rx, y - ry), (x + rx, y - ry), (x + rx, y + ry), (x - rx, y + ry))
        poly = Polygon(shell=path)
    else:
        # draw ellipse
        circ = Point(x, y).buffer(1)
        poly = scale(circ, rx, ry)  # type(ellipse)=polygon

    return poly



import numpy as np
from mcfnmr.demodata import case_dict
from mcfnmr.routines.utils import (
    prepareParameters,
    runMetaboMinerFits,
    cache_dir,
    makeResultsDF,
    totalWeights,
    collect_and_serialize_MM_grid_spectra,
)
from mcfnmr.routines.classification import classificationStats
from mcfnmr.demodata.loading import (
    loadMMMixtureLists,
    loadMMMixPointSpectra,
    loadMMLibrary,
)
from mcfnmr.routines.plotting import plotPerformanceScanResults, plotTimings
from mcfnmr.demodata import mix_selections


def makeScanTitle(caseSpec, smoothingRadius=0.0):
    mixIDs = caseSpec["mixIDs"]
    ptTargetStr = "pt" if caseSpec["pointTarget"] else "rst"
    if ptTargetStr == "rst" and smoothingRadius > 0:
        ptTargetStr += ", smooth%g" % smoothingRadius
    isolatedFitStr = "isofit" if caseSpec["isolated_fit"] else "jointfit"
    incrStr = ", incrFit" if caseSpec["incrementalFit"] else ""
    libID = caseSpec["libID"]
    fmttp = (str(mixIDs), ptTargetStr, incrStr, libID, isolatedFitStr)
    title = "mixes%s(%s)%s, lib %s(%s)" % fmttp
    return title


def runPerformanceScan(caseSpec, pars, thSpan, load=True):
    mixSubs = loadMMMixtureLists()
    lib = loadMMLibrary(pars["libID"])
    mixIDs = caseSpec["mixIDs"]
    if mixIDs == "all":
        mixIDs = mix_selections["SMART-Miner"]
    incrementalFit = caseSpec["incrementalFit"]

    if pars["pointTarget"]:
        mixSpecs = loadMMMixPointSpectra()
    else:
        pars["binning"] = None
        mixSpecs = collect_and_serialize_MM_grid_spectra(mixIDs, pars)

    if incrementalFit:
        results = runMetaboMinerFits(
            base_pars=pars,
            lib=lib,
            mixSpecs=mixSpecs,
            mixIDs=mixIDs,
            load=load,
            returnAllResults=True,
        )
        # Swap indexing order
        results = {
            ar: {mixID: results[mixID][ar] for mixID in mixIDs} for ar in pars["arSpan"]
        }

    else:
        results = dict()
        for ar in pars["arSpan"]:
            results[ar] = ({},)
            pars["assignmentRadius"] = ar
            results[ar] = runMetaboMinerFits(
                base_pars=pars, lib=lib, mixSpecs=mixSpecs, mixIDs=mixIDs, load=load
            )

    # Debug: timings plot
    # setup = caseSpec["name"]
    # plotTimings(results, setup, mixIDs, loglog=False, show=True)

    scanResults = {}
    for ar in pars["arSpan"]:
        dfs = {}
        scanResults[ar] = {}
        for mixID in mixIDs:
            result = results[ar][mixID]
            pars["targetSpecID"] = mixID
            dfTargets, dfCompounds = makeResultsDF(
                [{"res": result, "pars": pars}], originalWeights=totalWeights(lib)
            )
            dfs[mixID] = {
                "targets": dfTargets,
                "compounds": dfCompounds,
                "targetSpecID": pars["targetSpecID"],
                "targetTotalWeight": result.originalWeightY,
            }
        for th in thSpan:
            scores = classificationStats(dfs, pars, th, mixSubs, mixIDs, verb=False)
            scanResults[ar][th] = scores
    return scanResults


def main(cfg, show=False):
    assert cfg["task"] == "scan"
    assert cfg["lib"] == "MetaboMiner"
    assert cfg["mix"].split("_")[0] in ["N925", "N926", "N987", "N988", "N907", "all"]
    assert cfg["setup"] in ["A", "B", "C", "D", "E", "F"]

    setup = cfg["setup"]
    target_info = cfg["mix"].split("_")
    if len(target_info) > 1:
        smoothingRadius = float(target_info[2])
    else:
        smoothingRadius = 0.0
    mixID = target_info[0]

    setupSpec = case_dict[setup][mixID]
    incrementalFit = setupSpec["incrementalFit"]

    kwargs = dict(
        incrementalFit=incrementalFit,
        isolated_fit=setupSpec["isolated_fit"],
        noiseFilterLevel=None,
        pointTarget=setupSpec["pointTarget"],
        smoothing=smoothingRadius,
        workingDir=cfg["outdir"],
        recompute=cfg.get("recompute", True),
        # assignment radii to be scanned
        arSpan=np.array(sorted(cfg["ar"])),
    )

    pars = prepareParameters(
        specDomain="metabominer", targetSpecID=cfg["mix"], libID=cfg["lib"], **kwargs
    )

    # Scan performance in dependence of assignment radius  and detection threshold
    thSpan = np.linspace(*cfg["th"], cfg["nth"])
    scanResults = runPerformanceScan(
        setupSpec, pars, thSpan, load=not pars["recompute"]
    )
    axDict = plotPerformanceScanResults(
        scanResults,
        title=makeScanTitle(setupSpec, smoothingRadius),
        logAR=False,
        show=show,
    )


if __name__ == "__main__":
    # Test run
    cfg = dict(
        lib="MetaboMiner",
        # mix="N926_smooth_0.3",
        # mix="N925",
        mix="all",
        recompute=False,
        task="scan",
        setup="D",
        ar=[0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1, 0.12, 0.15],
        # ar=[0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1, 0.12, 0.15, 0.175, 0.2, 0.25],
        th=[0.0, 0.0015],
        nth=31,
        plot=True,
        outdir=cache_dir,
    )
    main(cfg, show=True)
